"""
fs.mountfs
==========

Contains MountFS class which is a virtual filesystem which can have other filesystems linked as branched directories.

For example, lets say we have two filesystems containing config files and resources respectively::

   [config_fs]
   |-- config.cfg
   `-- defaults.cfg

   [resources_fs]
   |-- images
   |   |-- logo.jpg
   |   `-- photo.jpg
   `-- data.dat

We can combine these filesystems in to a single filesystem with the following code::

    from fs.mountfs import MountFS
    combined_fs = MountFS()
    combined_fs.mountdir('config', config_fs)
    combined_fs.mountdir('resources', resources_fs)

This will create a single filesystem where paths under `config` map to `config_fs`, and paths under `resources` map to `resources_fs`::

    [combined_fs]
    |-- config
    |   |-- config.cfg
    |   `-- defaults.cfg
    `-- resources
        |-- images
        |   |-- logo.jpg
        |   `-- photo.jpg
        `-- data.dat

Now both filesystems can be accessed with the same path structure::

    print combined_fs.getcontents('/config/defaults.cfg')
    read_jpg(combined_fs.open('/resources/images/logo.jpg')

"""

from fs.base import *
from fs.errors import *
from fs.path import *
from fs import _thread_synchronize_default
from fs import iotools


class DirMount(object):
    def __init__(self, path, fs):
        self.path = path
        self.fs = fs

    def __str__(self):
        return "<DirMount %s, %s>" % (self.path, self.fs)

    def __repr__(self):
        return "<DirMount %s, %s>" % (self.path, self.fs)

    def __unicode__(self):
        return u"<DirMount %s, %s>" % (self.path, self.fs)


class FileMount(object):
    def __init__(self, path, open_callable, info_callable=None):
        self.open_callable = open_callable
        def no_info_callable(path):
            return {}
        self.info_callable = info_callable or no_info_callable


class MountFS(FS):
    """A filesystem that delegates to other filesystems."""

    _meta = { 'virtual': True,
              'read_only' : False,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              }

    DirMount = DirMount
    FileMount = FileMount

    def __init__(self, auto_close=True, thread_synchronize=_thread_synchronize_default):
        self.auto_close = auto_close
        super(MountFS, self).__init__(thread_synchronize=thread_synchronize)
        self.mount_tree = PathMap()

    def __str__(self):
        return "<%s [%s]>" % (self.__class__.__name__,self.mount_tree.items(),)

    __repr__ = __str__

    def __unicode__(self):
        return u"<%s [%s]>" % (self.__class__.__name__,self.mount_tree.items(),)

    def _delegate(self, path):
        path = abspath(normpath(path))
        object = None
        head_path = "/"
        tail_path = path

        for prefix in recursepath(path):
            try:
                object = self.mount_tree[prefix]
            except KeyError:
                pass
            else:
                head_path = prefix
                tail_path = path[len(head_path):]

        if type(object) is MountFS.DirMount:
            return object.fs, head_path, tail_path

        if type(object) is MountFS.FileMount:
            return self, "/", path

        try:
            self.mount_tree.iternames(path).next()
        except StopIteration:
            return None, None, None
        else:
            return self, "/", path

    @synchronize
    def close(self):
        # Explicitly closes children if requested
        if self.auto_close:
            for mount in self.mount_tree.itervalues():
                mount.fs.close()
        # Free references (which may incidently call the close method of the child filesystems)
        self.mount_tree.clear()
        super(MountFS, self).close()

    def getsyspath(self, path, allow_none=False):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            if allow_none:
                return None
            else:
                raise NoSysPathError(path=path)
        return fs.getsyspath(delegate_path, allow_none=allow_none)

    def getpathurl(self, path, allow_none=False):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            if allow_none:
                return None
            else:
                raise NoPathURLError(path=path)
        return fs.getpathurl(delegate_path, allow_none=allow_none)

    @synchronize
    def desc(self, path):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self:
            if fs.isdir(path):
                return "Mount dir"
            else:
                return "Mounted file"
        return "Mounted dir, maps to path %s on %s" % (abspath(delegate_path) or '/', str(fs))

    @synchronize
    def isdir(self, path):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            path = normpath(path)
            if path in ("/", ""):
                return True
            return False
        if fs is self:
            obj = self.mount_tree.get(path, None)
            return not isinstance(obj, MountFS.FileMount)
        return fs.isdir(delegate_path)

    @synchronize
    def isfile(self, path):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            return False
        if fs is self:
            obj = self.mount_tree.get(path, None)
            return isinstance(obj, MountFS.FileMount)
        return fs.isfile(delegate_path)

    @synchronize
    def exists(self, path):
        if path in ("/", ""):
            return True
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            return False
        if fs is self:
            return True
        return fs.exists(delegate_path)

    @synchronize
    def listdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is None:
            if path in ("/", ""):
                return []
            raise ResourceNotFoundError("path")

        elif fs is self:
            paths = self.mount_tree.names(path)
            return self._listdir_helper(path,
                                        paths,
                                        wildcard,
                                        full,
                                        absolute,
                                        dirs_only,
                                        files_only)
        else:
            paths = fs.listdir(delegate_path,
                               wildcard=wildcard,
                               full=False,
                               absolute=False,
                               dirs_only=dirs_only,
                               files_only=files_only)
            for nm in self.mount_tree.names(path):
                if nm not in paths:
                    if dirs_only:
                        if self.isdir(pathjoin(path,nm)):
                            paths.append(nm)
                    elif files_only:
                        if self.isfile(pathjoin(path,nm)):
                            paths.append(nm)
                    else:
                        paths.append(nm)
            if full or absolute:
                if full:
                    path = relpath(normpath(path))
                else:
                    path = abspath(normpath(path))
                paths = [pathjoin(path, p) for p in paths]

            return paths

    @synchronize
    def ilistdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is None:
            if path in ("/", ""):
                return
            raise ResourceNotFoundError(path)

        if fs is self:
            paths = self.mount_tree.names(path)
            for path in self._listdir_helper(path,paths,wildcard,full,absolute,dirs_only,files_only):
                yield path
        else:
            paths = fs.ilistdir(delegate_path,
                                wildcard=wildcard,
                                full=False,
                                absolute=False,
                                dirs_only=dirs_only)
            extra_paths = set(self.mount_tree.names(path))
            if full:
                pathhead = relpath(normpath(path))
                def mkpath(p):
                    return pathjoin(pathhead,p)
            elif absolute:
                pathhead = abspath(normpath(path))
                def mkpath(p):
                    return pathjoin(pathhead,p)
            else:
                def mkpath(p):
                    return p
            for p in paths:
                if p not in extra_paths:
                    yield mkpath(p)
            for p in extra_paths:
                if dirs_only:
                    if self.isdir(pathjoin(path,p)):
                        yield mkpath(p)
                elif files_only:
                    if self.isfile(pathjoin(path,p)):
                        yield mkpath(p)
                else:
                    yield mkpath(p)

    @synchronize
    def makedir(self, path, recursive=False, allow_recreate=False):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            raise UnsupportedError("make directory", msg="Can only makedir for mounted paths")
        if not delegate_path:
            if allow_recreate:
                return
            else:
                raise DestinationExistsError(path, msg="Can not create a directory that already exists (try allow_recreate=True): %(path)s")
        return fs.makedir(delegate_path, recursive=recursive, allow_recreate=allow_recreate)

    @synchronize
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        obj = self.mount_tree.get(path, None)
        if type(obj) is MountFS.FileMount:
            callable = obj.open_callable
            return callable(path, mode, **kwargs)

        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is self or fs is None:
            raise ResourceNotFoundError(path)

        return fs.open(delegate_path, mode, **kwargs)

    @synchronize
    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=64*1024):
        obj = self.mount_tree.get(path, None)
        if type(obj) is MountFS.FileMount:
            return super(MountFS, self).setcontents(path,
                                                    data,
                                                    encoding=encoding,
                                                    errors=errors,
                                                    chunk_size=chunk_size)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            raise ParentDirectoryMissingError(path)
        return fs.setcontents(delegate_path, data, encoding=encoding, errors=errors, chunk_size=chunk_size)

    @synchronize
    def createfile(self, path, wipe=False):
        obj = self.mount_tree.get(path, None)
        if type(obj) is MountFS.FileMount:
            return super(MountFS, self).createfile(path, wipe=wipe)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            raise ParentDirectoryMissingError(path)
        return fs.createfile(delegate_path, wipe=wipe)

    @synchronize
    def remove(self, path):
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            raise UnsupportedError("remove file", msg="Can only remove paths within a mounted dir")
        return fs.remove(delegate_path)

    @synchronize
    def removedir(self, path, recursive=False, force=False):
        path = normpath(path)
        if path in ('', '/'):
            raise RemoveRootError(path)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is self or fs is None:
            raise ResourceInvalidError(path, msg="Can not removedir for an un-mounted path")
        return fs.removedir(delegate_path, recursive, force)

    @synchronize
    def rename(self, src, dst):
        fs1, _mount_path1, delegate_path1 = self._delegate(src)
        fs2, _mount_path2, delegate_path2 = self._delegate(dst)

        if fs1 is not fs2:
            raise OperationFailedError("rename resource", path=src)

        if fs1 is not self:
            return fs1.rename(delegate_path1, delegate_path2)

        object = self.mount_tree.get(src, None)
        _object2 = self.mount_tree.get(dst, None)

        if object is None:
            raise ResourceNotFoundError(src)

        raise UnsupportedError("rename resource", path=src)

    @synchronize
    def move(self,src,dst,**kwds):
        fs1, _mount_path1, delegate_path1 = self._delegate(src)
        fs2, _mount_path2, delegate_path2 = self._delegate(dst)
        if fs1 is fs2 and fs1 is not self:
            fs1.move(delegate_path1,delegate_path2,**kwds)
        else:
            super(MountFS,self).move(src,dst,**kwds)

    @synchronize
    def movedir(self,src,dst,**kwds):
        fs1, _mount_path1, delegate_path1 = self._delegate(src)
        fs2, _mount_path2, delegate_path2 = self._delegate(dst)
        if fs1 is fs2 and fs1 is not self:
            fs1.movedir(delegate_path1,delegate_path2,**kwds)
        else:
            super(MountFS,self).movedir(src,dst,**kwds)

    @synchronize
    def copy(self,src,dst,**kwds):
        fs1, _mount_path1, delegate_path1 = self._delegate(src)
        fs2, _mount_path2, delegate_path2 = self._delegate(dst)
        if fs1 is fs2 and fs1 is not self:
            fs1.copy(delegate_path1,delegate_path2,**kwds)
        else:
            super(MountFS,self).copy(src,dst,**kwds)

    @synchronize
    def copydir(self,src,dst,**kwds):
        fs1, _mount_path1, delegate_path1 = self._delegate(src)
        fs2, _mount_path2, delegate_path2 = self._delegate(dst)
        if fs1 is fs2 and fs1 is not self:
            fs1.copydir(delegate_path1,delegate_path2,**kwds)
        else:
            super(MountFS,self).copydir(src,dst,**kwds)

    @synchronize
    def mountdir(self, path, fs):
        """Mounts a host FS object on a given path.

        :param path: A path within the MountFS
        :param fs: A filesystem object to mount

        """
        path = abspath(normpath(path))
        self.mount_tree[path] = MountFS.DirMount(path, fs)
    mount = mountdir

    @synchronize
    def mountfile(self, path, open_callable=None, info_callable=None):
        """Mounts a single file path.

        :param path: A path within the MountFS
        :param open_callable: A callable that returns a file-like object,
            `open_callable` should have the same signature as :py:meth:`~fs.base.FS.open`
        :param info_callable: A callable that returns a dictionary with information regarding the file-like object,
            `info_callable` should have the same signagture as :py:meth:`~fs.base.FS.getinfo`

        """
        self.mount_tree[path] = MountFS.FileMount(path, open_callable, info_callable)

    @synchronize
    def unmount(self, path):
        """Unmounts a path.

        :param path: Path to unmount
        :return: True if a path was unmounted, False if the path was already unmounted
        :rtype: bool

        """
        try:
            del self.mount_tree[path]
        except KeyError:
            return False
        else:
            return True

    @synchronize
    def settimes(self, path, accessed_time=None, modified_time=None):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is None:
            raise ResourceNotFoundError(path)

        if fs is self:
            raise UnsupportedError("settimes")
        fs.settimes(delegate_path, accessed_time, modified_time)

    @synchronize
    def getinfo(self, path):
        path = normpath(path)

        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is None:
            if path in ("/", ""):
                return {}
            raise ResourceNotFoundError(path)

        if fs is self:
            if self.isfile(path):
                return self.mount_tree[path].info_callable(path)
            return {}
        return fs.getinfo(delegate_path)

    @synchronize
    def getsize(self, path):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)

        if fs is None:
            raise ResourceNotFoundError(path)

        if fs is self:
            object = self.mount_tree.get(path, None)

            if object is None:
                raise ResourceNotFoundError(path)
            if not isinstance(object,MountFS.FileMount):
                raise ResourceInvalidError(path)

            size = object.info_callable(path).get("size", None)
            return size

        return fs.getinfo(delegate_path).get("size", None)

    @synchronize
    def getxattr(self,path,name,default=None):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            if path in ("/", ""):
                return default
            raise ResourceNotFoundError(path)
        if fs is self:
            return default
        return fs.getxattr(delegate_path,name,default)

    @synchronize
    def setxattr(self,path,name,value):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            raise ResourceNotFoundError(path)
        if fs is self:
            raise UnsupportedError("setxattr")
        return fs.setxattr(delegate_path,name,value)

    @synchronize
    def delxattr(self,path,name):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            raise ResourceNotFoundError(path)
        if fs is self:
            return True
        return fs.delxattr(delegate_path, name)

    @synchronize
    def listxattrs(self,path):
        path = normpath(path)
        fs, _mount_path, delegate_path = self._delegate(path)
        if fs is None:
            raise ResourceNotFoundError(path)
        if fs is self:
            return []
        return fs.listxattrs(delegate_path)





#!/usr/bin/env python

from netCDF4 import Dataset
from post_gnome import nc_particles
import shapefile as shp
import os
import zipfile
import numpy as np
import scipy.stats as st
from matplotlib.pyplot import contour


def write_proj_file(prj_filename):
    prj = open("%s.prj" % prj_filename, "w")
    epsg = 'GEOGCS["WGS 84",'
    epsg += 'DATUM["WGS_1984",'
    epsg += 'SPHEROID["WGS 84",6378137,298.257223563]]'
    epsg += ',PRIMEM["Greenwich",0],'
    epsg += 'UNIT["degree",0.0174532925199433]]'
    prj.write(epsg)
    prj.close()


def points(fn, package_dir, t2convert):

    nc = Dataset(fn)
    particles = nc_particles.Reader(nc)
    times = particles.times
    dt = [np.abs(((output_t - t2convert).total_seconds()) / 3600) for output_t in times]
    t = dt.index(min(dt))
    print 'Converting output from: ', times[t]

    w = shp.Writer(shp.POINT)
    w.autobalance = 1

    w.field('Year', 'C')
    w.field('Month', 'C')
    w.field('Day', 'C')
    w.field('Hour', 'C')
    w.field('LE id', 'N')
    w.field('Depth', 'N')
    w.field('Mass', 'N')
    w.field('Age', 'N')
    w.field('Status_Code', 'N')

    TheData = particles.get_timestep(t, variables=['latitude',
                                                   'longitude',
                                                   'id',
                                                   'depth',
                                                   'mass',
                                                   'age',
                                                   'status_codes'])
    for k, p in enumerate(zip(TheData['longitude'], TheData['latitude'])):
        w.point(p[0],p[1])
        w.record(times[t].year,
                 times[t].month,
                 times[t].day,
                 times[t].hour,
                 TheData['id'][k],
                 TheData['depth'][k],
                 TheData['mass'][k],
                 TheData['age'][k],
                 TheData['status_codes'][k])

    source_fdir = os.path.join(package_dir, 'source_files')
    shapefile_name = os.path.split(fn)[-1].split('.')[0]
    w.save(os.path.join(source_fdir, shapefile_name))

    nc.close()

    # create the PRJ file
    prj_filename = os.path.join(source_fdir, shapefile_name)
    write_proj_file(prj_filename)

    files = os.listdir(source_fdir)
    zipfname = shapefile_name + '.zip'
    zipf = zipfile.ZipFile(os.path.join(source_fdir, zipfname), 'w')
    for f in files:
        if f.split('.')[0] == shapefile_name:
            zipf.write(os.path.join(source_fdir, f), arcname=f)
    zipf.close()

    return zipfname


def contours(fn,
             package_dir,
             t2convert,
             levels=[0.1, 0.4, 0.8],
             names=['Light', 'Medium', 'Heavy'],
             ):

    print "contouring data in:", fn
    nc = Dataset(fn)
    particles = nc_particles.Reader(nc)
    times = particles.times
    dt = [np.abs(((output_t - t2convert).total_seconds()) / 3600) for output_t in times]
    t = dt.index(min(dt))
    print 'Converting output from: ', times[t]

    TheData = particles.get_timestep(t, variables=['latitude',
                                                   'longitude',
                                                   'id',
                                                   'depth',
                                                   'mass',
                                                   'age',
                                                   'status_codes'])

    # contouring
    status = TheData['status_codes']
    floating = np.where(status == 2)[0]
    x = TheData['longitude'][floating]
    y = TheData['latitude'][floating]

    # Peform the kernel density estimate
    xx, yy = np.mgrid[min(x) - .1:max(x) + .1:100j, min(y) - .1:max(y) + .1:100j]
    positions = np.vstack([xx.ravel(), yy.ravel()])
    values = np.vstack([x, y])
    kernel = st.gaussian_kde(values)
    f = np.reshape(kernel(positions).T, xx.shape)
    max_density = f.max()

    levels.sort()
    particle_contours = [lev * max_density for lev in levels]

    cs = contour(xx, yy, f, particle_contours)

    w = shp.Writer(shp.POLYGON)
    w.autobalance = 1

    w.field('Year', 'C')
    w.field('Month', 'C')
    w.field('Day', 'C')
    w.field('Hour', 'C')
    w.field('Depth', 'N')
    w.field('Type', 'C')

    for c in range(len(cs.collections)):
        p = cs.collections[c].get_paths()[0]
        v = p.vertices
        coords = [[[i[0], i[1]] for i in v]]
        w.poly(shapeType=3, parts=coords)
        w.record(times[t].year, times[t].month, times[t].day, times[t].hour,
                 TheData['depth'][c], names[c])
        print names[c]

    source_fdir = os.path.join(package_dir, 'source_files')
    shapefile_name = os.path.split(fn)[-1].split('.')[0]
    w.save(os.path.join(source_fdir, shapefile_name))

    nc.close()

    # create the PRJ file
    prj_filename = os.path.join(source_fdir, shapefile_name)
    write_proj_file(prj_filename)

    files = os.listdir(source_fdir)
    zipfname = shapefile_name + '.zip'
    zipf = zipfile.ZipFile(os.path.join(source_fdir, zipfname), 'w')
    for f in files:
        if f.split('.')[0] == shapefile_name:
            zipf.write(os.path.join(source_fdir, f), arcname=f)
    zipf.close()

    return zipfname



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
"""This module adds a fix for wx.py's introspect module.

In order to do code-completion, the function
`introspect.getAttributeName` accesses all the attributes of the
current object.  This causes severe problems for modules like tvtk
which depend on lazy importing.  The original introspect module also
has severe problems with large Numeric arrays because it calls str()
on the Numeric object in order to find its methods.

This file defines a fixed function that works fine with lazy objects
and large Numeric arrays.  This fixed function is injected into the
introspect module.
"""

# Import introspect.
from wx.py import introspect

import types

# The fixed function.
def getAttributeNames(object, includeMagic=1, includeSingle=1,
                      includeDouble=1):
    """Return list of unique attributes, including inherited, for object."""
    attributes = []
    dict = {}
    if not introspect.hasattrAlwaysReturnsTrue(object):
        # Add some attributes that don't always get picked up.
        special_attrs = ['__bases__', '__class__', '__dict__', '__name__',
                         'func_closure', 'func_code', 'func_defaults',
                         'func_dict', 'func_doc', 'func_globals', 'func_name']
        attributes += [attr for attr in special_attrs \
                       if hasattr(object, attr)]
    # For objects that have traits, get all the trait names since
    # these do not show up in dir(object).
    if hasattr(object, 'trait_names'):
        try:
            attributes += object.trait_names()
        except TypeError:
            pass

    if includeMagic:
        try: attributes += object._getAttributeNames()
        except: pass
    # Get all attribute names.
    attrdict = getAllAttributeNames(object)
    # Store the object's dir.
    object_dir = dir(object)
    for (obj_type_name, technique, count), attrlist in attrdict.items():
        # This complexity is necessary to avoid accessing all the
        # attributes of the object.  This is very handy for objects
        # whose attributes are lazily evaluated.
        if type(object).__name__ == obj_type_name and technique == 'dir':
            attributes += attrlist
        else:
            attributes += [attr for attr in attrlist \
                           if attr not in object_dir and \
                           hasattr(object, attr)]

    # Remove duplicates from the attribute list.
    for item in attributes:
        dict[item] = None
    attributes = dict.keys()
    # new-style swig wrappings can result in non-string attributes
    # e.g. ITK http://www.itk.org/
    attributes = [attribute for attribute in attributes \
                  if isinstance(attribute, basestring)]
    attributes.sort(lambda x, y: cmp(x.upper(), y.upper()))
    if not includeSingle:
        attributes = filter(lambda item: item[0]!='_' \
                            or item[1]=='_', attributes)
    if not includeDouble:
        attributes = filter(lambda item: item[:2]!='__', attributes)
    return attributes


# Replace introspect's version with ours.
introspect.getAttributeNames = getAttributeNames

# This is also a modified version of the function which does not use
# str(object).
def getAllAttributeNames(object):
    """Return dict of all attributes, including inherited, for an object.

    Recursively walk through a class and all base classes.
    """
    attrdict = {}  # (object, technique, count): [list of attributes]
    # !!!
    # Do Not use hasattr() as a test anywhere in this function,
    # because it is unreliable with remote objects: xmlrpc, soap, etc.
    # They always return true for hasattr().
    # !!!
    try:
        # This could(?) fail if the type is poorly defined without
        # even a name.
        key = type(object).__name__
    except:
        key = 'anonymous'
    # Wake up sleepy objects - a hack for ZODB objects in "ghost" state.
    wakeupcall = dir(object)
    del wakeupcall
    # Get attributes available through the normal convention.
    attributes = dir(object)
    attrdict[(key, 'dir', len(attributes))] = attributes
    # Get attributes from the object's dictionary, if it has one.
    try:
        attributes = object.__dict__.keys()
        attributes.sort()
    except:  # Must catch all because object might have __getattr__.
        pass
    else:
        attrdict[(key, '__dict__', len(attributes))] = attributes
    # For a class instance, get the attributes for the class.
    try:
        klass = object.__class__
    except:  # Must catch all because object might have __getattr__.
        pass
    else:
        if klass is object:
            # Break a circular reference. This happens with extension
            # classes.
            pass
        else:
            attrdict.update(getAllAttributeNames(klass))
    # Also get attributes from any and all parent classes.
    try:
        bases = object.__bases__
    except:  # Must catch all because object might have __getattr__.
        pass
    else:
        if isinstance(bases, types.TupleType):
            for base in bases:
                if type(base) is types.TypeType:
                    # Break a circular reference. Happens in Python 2.2.
                    pass
                else:
                    attrdict.update(getAllAttributeNames(base))
    return attrdict



import os
import sys
import datetime

from pyface.api import ImageResource

# Major package imports.
from fs.opener import Opener, OpenerRegistry, _FSClosingFile, opener, fsopen
from fs.base import FS
from fs.path import normpath
import fs
import wx

from omnivore.templates import get_template_fh, find_template_path

import logging
log = logging.getLogger(__name__)

from omnivore import __version__, __author__, __author_email__, __url__
substitutes = {
    'prog': 'Omnivore XL',
    'yearrange': '2014-2017',
    'version': __version__,
    'description': "The binary editor and disassembler for modern 8-bit hackers",
    'tagline': "Byte into the meat of 8-bit software!",
    'author': __author__,
    'author_email': __author_email__,
    'url': __url__,
    'license': "Application licensed under the GNU General Public License, version 3.0",
    'warning': """<P>This code is still under development, so make backups and save often.""",
    'license_text': """This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
""",
    'authors': "",
    'contributors': """<p>Contributors to Omnivore:</p>
<ul>
<li>Mark James for the <a href=\"http://www.famfamfam.com/lab/icons/silk/\">free silk icon set</a>
<li>Chris Barker for bug reports and fixes on OS X</a>
<li>Kevin Savetz of <a href=\"http://ataripodcast.com\">ANTIC, the Atari 8-bit Podcast</a> for beta testing
<li>Wade Ripkowski of <a href=\"http://inverseatascii.info\">Inverse ATASCII, the Atari 8-bit Productivity Podcast</a> for beta testing
<li>David Beazley, author of <a href=\"http://dabeaz.com\">Python Cookbook, 3rd Ed. and more</a>, for the 6502 mini assembler
<li>Jeff Tranter, for <a href=\"https://github.com/robmcmullen/udis\">udis</a>, the Universal Disassembler for 8-bit microprocessors
<li>Charles Mangin for the Apple ][ and KIM-1 memory maps
</ul>""",
    }
substitutes['copyright'] = 'Copyright (c) %(yearrange)s %(author)s (%(author_email)s)' % substitutes

about = {
    "omnivore": """<html>
<h2>%(prog)s %(version)s</h2>

<h3>%(description)s</h3>

<p><img src="about://logo.png">

<h5>%(tagline)s</h5>

<p>%(copyright)s</p>

<p>%(license)s</p>

<p>%(warning)s</p>

<p>%(contributors)s</p>
""" % substitutes,
    "logo.png": ImageResource('omnivore256'),
    }


# FIXME: can't seem to figure out how to create a subclass of InputStream,
# so can't create a new FSFile object directly from the file in the fs
# about: filesystem.  So, fake the about: filesystem for the HTML control
# by duplicating the file from the fs about: filesystem and passing that
# wx.FSFile object

#class MemoryInputStream(wx.InputStream):
#    def __init__(self, fh):
#        wx.InputStream.__init__(self)
#        self.fh = fh
#        self.cursor = 0
#
#    def OnSysRead(self, *args):
#        log.error("OnSysRead!!!!!")

class WxAboutFileSystemHandler(wx.FileSystemHandler):
    def CanOpen(self, location):
        return self.GetProtocol(location) == "about"

    def OpenFile(self, fs, location):
        # For some reason, the actual path shows up as the right location
        # rather than left, and it includes the leading slashes.
        path = self.GetRightLocation(location).lstrip("/")
        wxfs = wx.FileSystem()
        fsfile = wxfs.OpenFile("memory:%s" % path)
        if fsfile is None:
            try:
                fh = opener.open(location, "rb")
            except errors.FSError, e:
                log.error(str(e))
                return None
            log.debug("Created %s in wxMemoryFS" % path)
            wx.MemoryFSHandler.AddFile(path, fh.read())

            fsfile = wxfs.OpenFile("memory:%s" % path)
        else:
            log.debug("Found %s in wxMemoryFS" % path)
        return fsfile


def parse_with_about(self, fs_url, default_fs_name=None, writeable=False, create_dir=False, cache_hint=True):
    """Handle special "about://" URLs.  Normal parse function puts the path
    part of the about: url into the url2 component, not the path. E.g.
    
    >>> opener.split_segments("about://omnivore").groups()
    ('about', None, None, 'omnivore', None)
    
    Adding a bang character puts the subsequent chars in the path component:
    
    >>> opener.split_segments("about://omnivore!oeuae").groups()
    ('about', None, None, 'omnivore', 'oeuae')
    
    so putting the bang character immediately after the protocol causes the
    path to appear in the right place:
    
    >>> opener.split_segments("about://!omnivore!oeuae").groups()
    ('about', None, None, '', 'omnivore!oeuae')
    
    """
    match = self.split_segments(fs_url)
    if match:
        fs_name, credentials, url1, url2, path = match.groups()
        if fs_name == "about":
            # Force the path
            fs_url = "about://!%s" % url2
    return self.old_parse(fs_url, default_fs_name, writeable, create_dir, cache_hint)


OpenerRegistry.old_parse = OpenerRegistry.parse
OpenerRegistry.parse = parse_with_about


class AboutOpener(Opener):
    names = ['about']
    desc = """In-memory filesystem that holds files until the exiting the application

examples:
* about:// (opens a new memory filesystem)
* about://foo/bar (opens a new memory filesystem with subdirectory /foo/bar)
    """

    about_fs = None

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        if cls.about_fs is None:
            from fs.memoryfs import MemoryFS
            cls.about_fs = MemoryFS()
            cls.about_fs.read_only_fs = True
        memfs = cls.about_fs
        if create_dir:
            memfs = memfs.makeopendir(fs_path)
        return memfs, None


class TemplateFS(FS):
    """Simple fs to read files from the template directories.

    """

    _meta = {'read_only': True,
             'network': False}

    def __init__(self, url):
        self.root_url = url

    def _make_url(self, path):
        path = normpath(path)
        url = '%s/%s' % (self.root_url.rstrip('/'), path.lstrip('/'))
        return url

    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):

        if '+' in mode or 'w' in mode or 'a' in mode:
            raise UnsupportedError('write')

        url = self._make_url(path)
        log.debug("TemplateFS: loading %s" % url)
        try:
            fh = get_template_fh(url)
        except OSError, e:
            raise fs.errors.ResourceNotFoundError(path, details="No template found. %s" % e)

        return fh

    def exists(self, path):
        url = self._make_url(path)
        return get_template_fh(url) is not None

    def isdir(self, path):
        return False

    def isfile(self, path):
        return self.exists(path)

    def listdir(self, path="./",
                      wildcard=None,
                      full=False,
                      absolute=False,
                      dirs_only=False,
                      files_only=False):
        return []

    def getinfo(self, path):
        url = self._make_url(path)
        path = find_template_path(url)
        stats = self._stat(path)
        info = dict((k, getattr(stats, k)) for k in dir(stats) if k.startswith('st_'))
        info['size'] = info['st_size']
        #  TODO: this doesn't actually mean 'creation time' on unix
        fromtimestamp = datetime.datetime.fromtimestamp
        ct = info.get('st_ctime', None)
        if ct is not None:
            info['created_time'] = fromtimestamp(ct)
        at = info.get('st_atime', None)
        if at is not None:
            info['accessed_time'] = fromtimestamp(at)
        mt = info.get('st_mtime', None)
        if mt is not None:
            info['modified_time'] = fromtimestamp(mt)
        return info


class TemplateOpener(Opener):
    names = ['template']
    desc = """Simple filesystem that loads from the template directories

examples:
* about:// (opens a new memory filesystem)
* about://foo/bar (opens a new memory filesystem with subdirectory /foo/bar)
    """

    @classmethod
    def get_fs(cls, registry, fs_name, fs_name_params, fs_path,  writeable, create_dir):
        if '/' in fs_path:
            dirname, resourcename = fs_path.rsplit('/', 1)
        else:
            dirname = "/"
            resourcename = fs_path
        fs = TemplateFS(dirname)
        return fs, resourcename


def init_filesystems():
    wx.FileSystem.AddHandler(WxAboutFileSystemHandler())
    wx.FileSystem.AddHandler(wx.MemoryFSHandler())

    init_about_filesystem()
#
#    for name in about.iterkeys():
#        url = "about://%s" % name
#        fh = opener.open(url, "rb")
#        text = fh.read()
#        print "Found %s, length=%d" % (name, len(text))
#        match = opener.split_segments(url)
#        stuff = match.groups()
#        print "url=%s, stuff=%s" % (url, str(stuff))
#        fs, path = opener.parse(url)
#        print "fs=%s (%s), path=%s" % (fs, id(fs), path)


def init_about_filesystem():
    opener.add(AboutOpener)
    opener.add(TemplateOpener)
    for name, text in about.iteritems():
        url = "about://%s" % name
        fh = opener.open(url, "wb")
        if hasattr(text, 'create_bitmap'):
            fhb = opener.open(text.absolute_path, "rb")
            text = fhb.read()
            fhb.close()
        fh.write(text)
        fh.close()
        log.debug("Created %s" % url)



import wx


class EscapeCancelMixin(object):
    """Segment selector for choosing which portion of the binary data to view
    """

    def register_escape_handler(self):
            self.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)

    def on_char_hook(self, evt):
        if evt.GetKeyCode() == wx.WXK_ESCAPE:
            wx.CallAfter(self.task.on_hide_minibuffer_or_cancel, None)
        else:
            evt.Skip()



#-------------------------------------------------------------------------------
#
#  A simple dual scatter-plot template defined as a test for the template
#  package.
#
#  Written by: David C. Morrill
#  (based on the original cp.plot geo_scatter_plot.py file)
#
#  Date: 08/01/2007
#
#  (c) Copy 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A simple dual scatter-plot template defined as a test for the template
    package.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Undefined

from traitsui.api \
    import View, VGroup, Item, Label, Theme, TextEditor

from traitsui.wx.themed_slider_editor \
    import ThemedSliderEditor

from traitsui.wx.themed_text_editor \
    import ThemedTextEditor

from enable.api \
    import ColorTrait

from chaco.api \
    import HPlotContainer

from chaco.scatter_markers \
    import marker_trait

from apptools.template.api \
    import Template, TRange, TStr, TInstance, TDerived

from enable_editor \
    import EnableEditor

from scatter_plot \
    import ScatterPlot

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# Template color trait:
TColor = ColorTrait( template = 'copy' )

#-------------------------------------------------------------------------------
#  'ScatterPlot2' class:
#-------------------------------------------------------------------------------

class ScatterPlot2 ( Template ):

    #-- Template Traits --------------------------------------------------------

    # The title of the plot:
    title = TStr( 'Dual Scatter Plots' )

    # The type of marker to use.  This is a mapped trait using strings as the
    # keys:
    marker = marker_trait( template = 'copy', event = 'update' )

    # The pixel size of the marker (doesn't include the thickness of the
    # outline):
    marker_size = TRange( 1, 5, 1, event = 'update' )

    # The thickness, in pixels, of the outline to draw around the marker.  If
    # this is 0, no outline will be drawn.
    line_width = TRange( 0.0, 5.0, 1.0 )

    # The fill color of the marker:
    color = TColor( 'red', event = 'update' )

    # The color of the outline to draw around the marker
    outline_color = TColor( 'black', event = 'update' )

    # The amount of space between plots:
    spacing = TRange( 0.0, 20.0, 0.0 )

    # The contained scatter plots:
    scatter_plot_1 = TInstance( ScatterPlot, () )
    scatter_plot_2 = TInstance( ScatterPlot, () )

    #-- Derived Traits ---------------------------------------------------------

    plot = TDerived

    #-- Traits UI Views --------------------------------------------------------

    # The scatter plot view:
    template_view = View(
        VGroup(
            Item( 'title',
                  show_label = False,
                  style      = 'readonly',
                  editor     = ThemedTextEditor(
                                 theme = Theme( '@GBB', alignment = 'center' ) )
            ),
            Item( 'plot',
                  show_label = False,
                  resizable  = True,
                  editor     = EnableEditor(),
                  item_theme = Theme( '@GF5', margins = 0 )
            )
        ),
        resizable = True
    )

    # The scatter plot options view:
    options_view = View(
        VGroup(
            VGroup(
                Label( 'Scatter Plot Options',
                       item_theme = Theme( '@GBB', alignment = 'center' ) ),
                show_labels = False
            ),
            VGroup(
                Item( 'title', editor = TextEditor() ),
                Item( 'marker' ),
                Item( 'marker_size', editor = ThemedSliderEditor() ),
                Item( 'line_width',
                      label  = 'Line Width',
                      editor = ThemedSliderEditor() ),
                Item( 'spacing', editor = ThemedSliderEditor() ),
                Item( 'color',         label = 'Fill Color' ),
                Item( 'outline_color', label = 'Outline Color' ),
                group_theme = Theme( '@GF5', margins = ( -5, -1 ) ),
                item_theme  = Theme( '@G0B', margins = 0 )
            )
        )
    )

    #-- ITemplate Interface Implementation -------------------------------------

    def activate_template ( self ):
        """ Converts all contained 'TDerived' objects to real objects using the
            template traits of the object. This method must be overridden in
            subclasses.

            Returns
            -------
            None
        """
        plots = [ p for p in [ self.scatter_plot_1.plot,
                               self.scatter_plot_2.plot ] if p is not None ]
        if len( plots ) == 2:
            self.plot = HPlotContainer( spacing = self.spacing )
            self.plot.add( *plots )
        elif len( plots ) == 1:
            self.plot = plots[0]

    #-- Default Values ---------------------------------------------------------

    def _scatter_plot_1_default ( self ):
        """ Returns the default value for the first scatter plot.
        """
        result = ScatterPlot()
        result.index.description  = 'Shared Plot Index'
        result.value.description += ' 1'

        return result

    def _scatter_plot_2_default ( self ):
        """ Returns the default value for the second scatter plot.
        """
        result = ScatterPlot( index = self.scatter_plot_1.index )
        result.value.description += ' 2'
        result.value.optional = True

        return result

    #-- Trait Event Handlers ---------------------------------------------------

    def _update_changed ( self, name, old, new ):
        """ Handles a plot option being changed.
        """
        setattr( self.scatter_plot_1, name, new )
        setattr( self.scatter_plot_2, name, new )
        self.plot = Undefined

    def _spacing_changed ( self, spacing ):
        """ Handles the spacing between plots being changed.
        """
        self.plot = Undefined




#-----------------------------------------------------------------------------
# Name:        imagescroller.py
# Purpose:     scrolling container for bitmapped images
#
# Author:      Rob McMullen
#
# Created:     2007
# RCS-ID:      $Id: $
# Copyright:   (c) 2007-2016 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
"""ImageScroller -- a container for viewing bitmapped images.

This control is designed to be a generic bitmap viewer that can scroll
to handle large images.  In addition, user interaction with the bitmap
is possible through the use of subclasses of the MouseSelector class.
Currently, a crosshair selector and a rubber band selector are
provided, with others to follow.

Coordinate systems
==================

Event coords are in terms of a fixed viewport the size of the client
area of the window containing the scrolled window.  The scrolled
window itself has an origin that can be negative relative to this
viewport.

World coordinates are in terms of the size of canvas of the scrolled
window itself, not the smaller size of the viewport onto the scrolled
window.  The size of the scrolled window's canvas is determined by the
scaled size of the bitmap containing it.

Coordinate systems example
--------------------------

If the original bitmap is 100 pixels wide and 200 pixels high, and the
zoom factor is 4, the size of both the scaled image and the scrolled
window canvas will be 400 x 800 pixels.  If the viewport of the
scrolled window is 200 x 300, some of the scaled bitmap won't be
displayed.

 +-------------+
 |             |--- scaled bitmap
 |    +------+ |
 |    |      | |
 |    |      |----- scrolled window viewport at location (x, y) of the bitmap
 |    |      | |
 |    +------+ |
 |             |
 +-------------+

Again, world coordinates are in terms of the scaled bitmap, but event
coordinates are in terms of the viewport.  An event coordinate of (0,
0) actually occurs at (x, y) on the bitmap.

@author: Rob McMullen
@version: 0.5

Changelog:
    0.4:
        * added methods to allow programmatic specification of rubberband
        * added setSelection to Rubberband
        * added startSelector to ImageScroller
    0.5:
        * changed to wx.Overlay for drawing (instead of XOR)
"""

import os

import wx
import wx.lib.newevent

try:
    from peppy.debug import *
except:
    def dprint(txt=""):
        #print txt
        pass


##### - Here are some utility functions from wx.lib.mixins.rubberband

def normalizeBox(box):
    """
    Convert any negative measurements in the current
    box to positive, and adjust the origin.
    """
    x, y, w, h = box
    if w < 0:
        x += (w+1)
        w *= -1
    if h < 0:
        y += (h+1)
        h *= -1
    return (x, y, w, h)


def boxToExtent(box):
    """
    Convert a box specification to an extent specification.
    I put this into a seperate function after I realized that
    I had been implementing it wrong in several places.
    """
    b = normalizeBox(box)
    return (b[0], b[1], b[0]+b[2]-1, b[1]+b[3]-1)


def pointInBox(x, y, box):
    """
    Return True if the given point is contained in the box.
    """
    e = boxToExtent(box)
    state = x >= e[0] and x <= e[2] and y >= e[1] and y <= e[3]
    # dprint("x=%d y=%d box=%s state=%s" % (x, y, e, state))
    return state


def pointOnBox(x, y, box, thickness=1):
    """
    Return True if the point is on the outside edge
    of the box.  The thickness defines how thick the
    edge should be.  This is necessary for HCI reasons:
    For example, it's normally very difficult for a user
    to manuever the mouse onto a one pixel border.
    """
    outerBox = box
    innerBox = (box[0]+thickness, box[1]+thickness, box[2]-(thickness*2), box[3]-(thickness*2))
    return pointInBox(x, y, outerBox) and not pointInBox(x, y, innerBox)


class MouseSelector(object):
    """Base class for user interaction classes.

    A selector is a class that represents a user interaction with the
    bitmap.  User interactions are triggered by a certain mouse event
    combination, and once triggered, the ImageScroller makes that the
    current selector.

    Mouse event processing for the ImageScroller is directed to the
    current selector through the processEvent method, which directs
    the action depending on which combination of mouse events occur.
    The processEvent handler should return False when an event
    combination is reached that signals the end of the selector's
    processing.  The selector will be destroyed by the ImageScroller
    when this occurs.
    """
    cursor = wx.CURSOR_ARROW

    def __init__(self, scroller, ev=None):
        self.scroller = scroller
        self.world_coords = None
        self.start_img_coords = None
        self.last_img_coords = None

        # autoscrolling enabled?
        self.want_autoscroll = False

        # cursor stuff
        self.want_blank_cursor = False

        if ev is not None:
            self.startEvent(ev)

    @classmethod
    def trigger(self, ev):
        """Identify the trigger event to turn on this selector.

        Return True if the event passed in is the event that triggers
        this selector to begin.
        """
        if ev.LeftDown():
            return True
        return False

    def processEvent(self, ev):
        """Process a mouse event for this selector.

        This callback is called for any mouse event once the selector
        is active in the scroller.  The selector is not deactivated
        until this handler returns False, so make sure some event will
        cause it to return False.
        """
        if ev.LeftIsDown() and ev.Dragging():
            self.handleEvent(ev)
        elif ev.LeftUp():
            self.finishEvent(ev)
            return False
        return True

    def startEvent(self, ev):
        """Set up the initial state to begin event processing.

        This is generally called by the constructor to set up whatever
        initial state is necessary to allow further events to be
        handled by the processEvent method.
        """
        coords = self.scroller.convertEventCoords(ev)
        self.scroller.CaptureMouse()
        self.start_img_coords = self.scroller.getImageCoords(*coords)
        self.setWorldCoordsFromImageCoords(*self.start_img_coords)
        self.draw()
        self.handleEventPostHook(ev)
        # dprint("ev: (%d,%d), coords: (%d,%d)" % (ev.GetX(), ev.GetY(), coords[0], coords[1]))
        self.want_autoscroll = True

    def handleEvent(self, ev):
        """General event handler.

        This method is called by processEvent to show the results of
        the user interaction on the bitmap.
        """
        coords = self.scroller.convertEventCoords(ev)
        img_coords = self.scroller.getImageCoords(*coords)
        if img_coords != self.last_img_coords:
            self.setWorldCoordsFromImageCoords(*img_coords)
            self.draw()
            self.handleEventPostHook(ev)
        #dprint("ev: (%d,%d), coords: (%d,%d)" % (ev.GetX(), ev.GetY(), coords[0], coords[1]))

    def handleEventPostHook(self, ev):
        """Hook called during every event handled by L{handleEvent}
        
        This is a no-op in the base class; it's designed to be overridden in
        subclasses.
        """
        pass

    def finishEvent(self, ev):
        """User interaction is complete.

        This method is called when the user has completed the
        interaction.  This does not necessarily destroy the selector
        -- it's up to the processEvent returning False to signal the
        ImageScroller to kill the selector.

        This could be used in multi-stage processing; see the
        RubberBand for an example where finishEvent doesn't mean that
        the selector should be destroyed.
        """
        self.want_autoscroll = False
        self.scroller.ReleaseMouse()
        self.cleanup()
        self.finishEventPostHook(ev)

    def finishEventPostHook(self, ev):
        """Hook that gets called when the selector has been dismissed.
        
        This is a no-op in the base class; it's designed to be overridden in
        subclasses.
        """
        pass

    def draw(self):
        #print("draw")
        if self.start_img_coords:
            dc, odc = self.getOverlayDC()
            self.drawSelector(dc)
            del odc # workaround big in python wrappers; delete overlay dc first

    def cleanup(self):
        dc = wx.ClientDC(self.scroller)
        odc = wx.DCOverlay(self.scroller.overlay, dc)
        odc.Clear()
        del odc
        self.scroller.overlay.Reset()
        self.scroller.Refresh()
        self.world_coords = None

    def recalc(self):
        self.setWorldCoordsFromImageCoords(*self.last_img_coords)

    def recalc_and_draw(self):
        #print("recalc and draw")
        self.scroller.overlay.Reset()
        self.recalc()
        self.draw()

    def getOverlayDC(self):
        dc = wx.ClientDC(self.scroller)
        odc = wx.DCOverlay(self.scroller.overlay, dc)
        odc.Clear()
        dc.SetPen(wx.Pen(wx.WHITE, 1, wx.DOT))
        dc.SetBrush(wx.TRANSPARENT_BRUSH)
        return dc, odc

    def getViewOffset(self):
        xView, yView = self.scroller.GetViewStart()
        xDelta, yDelta = self.scroller.GetScrollPixelsPerUnit()
        xoff = xView * xDelta
        yoff = yView * yDelta
        return -xoff, -yoff

    def setWorldCoordsFromImageCoords(self, x, y):
        self.last_img_coords = (x, y)
        zoom = self.scroller.zoom
        offset = int(zoom / 2)
        x = int(x * zoom)
        y = int(y * zoom)
        self.world_coords = (x + offset, y + offset)


class NullSelector(MouseSelector):
    def startEvent(self, ev):
        pass

    def handleEvent(self, ev):
        pass

    def draw(self):
        pass


# create a new Event class and a EVT binder function for a crosshair
# motion event
(CrosshairMotionEvent, EVT_CROSSHAIR_MOTION) = wx.lib.newevent.NewEvent()


class Crosshair(MouseSelector):
    def __init__(self, scroller, ev=None):
        MouseSelector.__init__(self, scroller)

        self.want_blank_cursor = True
        self.crossbox = None

        if ev is not None:
            self.startEvent(ev)

    def handleEventPostHook(self, ev):
        x, y = self.last_img_coords
        if self.scroller.crop is not None:
            dx, dy, w, h = self.scroller.crop
        else:
            dx, dy = (0, 0)
        cev = CrosshairMotionEvent(imageCoords = (x, y),
                                   uncroppedImageCoords = (x + dx, y + dy))
        wx.PostEvent(self.scroller, cev)

    def drawCrossBox(self, x, y, xoff, yoff, dc):
        dc.DrawRectangle(self.crossbox[0] + xoff, self.crossbox[1] + yoff,
                         self.crossbox[2], self.crossbox[3])
        dc.DrawLine(x, 0, x, self.crossbox[1] + yoff)
        dc.DrawLine(x, self.crossbox[1] + self.crossbox[3] + yoff + 1,
                    x, self.scroller.height)
        dc.DrawLine(0, y, self.crossbox[0] + xoff, y)
        dc.DrawLine(self.crossbox[0] + self.crossbox[2] + xoff + 1, y,
                    self.scroller.width, y)

    def drawSelector(self, dc):
        xoff, yoff = self.getViewOffset()
        x = self.world_coords[0] + xoff
        y = self.world_coords[1] + yoff
        if self.crossbox:
            dc.SetPen(wx.Pen(wx.BLACK, 1))
            self.drawCrossBox(x, y, xoff, yoff, dc)
            dc.SetPen(wx.Pen(wx.WHITE, 1, wx.DOT))
            self.drawCrossBox(x, y, xoff, yoff, dc)
        else:
            dc.SetPen(wx.Pen(wx.BLACK, 1))
            dc.DrawLine(x, 0, x, self.scroller.height)
            dc.DrawLine(0, y, self.scroller.width, y)
            dc.SetPen(wx.Pen(wx.WHITE, 1, wx.DOT))
            dc.DrawLine(x, 0, x, self.scroller.height)
            dc.DrawLine(0, y, self.scroller.width, y)

    def setWorldCoordsFromImageCoords(self, x, y):
        self.last_img_coords = (x, y)
        zoom = self.scroller.zoom
        offset = int(zoom / 2)
        x = int(x * zoom)
        y = int(y * zoom)
        self.world_coords = (x + offset, y + offset)
        if self.scroller.zoom >= 1:
            self.crossbox = (x-1, y-1, zoom + 2, zoom + 2)
        else:
            self.crossbox = None
        # dprint("crosshair = %s, img = %s" % (self.world_coords, self.last_img_coords))


# create a new Event class and a EVT binder function for a crosshair
# motion event
(RubberBandSizeEvent, EVT_RUBBERBAND_SIZE) = wx.lib.newevent.NewEvent()
(RubberBandMotionEvent, EVT_RUBBERBAND_MOTION) = wx.lib.newevent.NewEvent()


class RubberBand(MouseSelector):
    move_cursor = wx.CURSOR_SIZING
    resize_cursors = [wx.CURSOR_SIZENWSE,
                      wx.CURSOR_SIZENS,
                      wx.CURSOR_SIZENESW,
                      wx.CURSOR_SIZEWE,
                      wx.CURSOR_SIZENWSE,
                      wx.CURSOR_SIZENS,
                      wx.CURSOR_SIZENESW,
                      wx.CURSOR_SIZEWE
                      ]

    def __init__(self, scroller, ev=None):
        MouseSelector.__init__(self, scroller)

        self.border_sensitivity = 3
        self.resize_index = None
        self.event_type = None
        self.move_img_coords = None

        if ev is not None:
            self.startEvent(ev)

    def processEvent(self, ev):
        """Process a mouse event for this selector.

        This callback is called for any mouse event once the selector
        is active in the scroller.  The selector is not deactivated
        until this handler returns False, so make sure some event will
        cause it to return False.
        """
        if ev.LeftDown():
            # restart event if we get another LeftDown
            self.startEvent(ev)
        elif ev.LeftIsDown() and ev.Dragging():
            self.handleEvent(ev)
        elif ev.LeftUp():
            self.finishEvent(ev)
        elif ev.Moving():
            # no mouse buttons; change cursor if over resize box
            self.handleCursorChanges(ev)
        return True

    def startEvent(self, ev):
        """Driver for new event.

        This selector recognizes a few different types of events: a
        normal event where the user uses the mouse to select a new
        rectangular area, a move event where the user can drag around
        the area without changing its size, and a bunch of resize
        events where the user can grab a corner or edge and make the
        rectangular area bigger.
        """
        self.scroller.CaptureMouse()
        coords = self.scroller.convertEventCoords(ev)
        # dprint("mouse=%s world=%s" % (coords, self.world_coords))
        if self.isOnBorder(coords):
            self.startResizeEvent(ev, coords)
        elif self.isInside(coords):
            self.startMoveEvent(ev, coords)
        else:
            self.startNormalEvent(ev, coords)
        self.want_autoscroll = True

    def startNormalEvent(self, ev, coords):
        self.event_type = None
        self.start_img_coords = self.scroller.getImageCoords(*coords)
        self.setWorldCoordsFromImageCoords(*self.start_img_coords)
        self.draw()
        self.handleEventPostHook(ev)
        # dprint("ev: (%d,%d), coords: (%d,%d)" % (ev.GetX(), ev.GetY(), coords[0], coords[1]))

    def startResizeEvent(self, ev, coords):
        self.event_type = "resize"
        self.normalizeImageCoords()
        self.move_img_coords = self.scroller.getImageCoords(*coords)
        # dprint("%s: index=%d starting from %s" % (self.event_type, self.resize_index, self.move_img_coords))

    def startMoveEvent(self, ev, coords):
        self.event_type = "move"
        self.normalizeImageCoords()
        self.move_img_coords = self.scroller.getImageCoords(*coords)
        # dprint("%s: starting from %s" % (self.event_type, self.move_img_coords))

    def handleEvent(self, ev):
        if self.event_type == "resize":
            self.handleResizeEvent(ev)
        elif self.event_type == "move":
            self.handleMoveEvent(ev)
        else:
            MouseSelector.handleEvent(self, ev)
        # dprint(self.world_coords)

    def handleResizeEvent(self, ev):
        coords = self.scroller.convertEventCoords(ev)
        img_coords = self.scroller.getImageCoords(coords[0],
                                                  coords[1],
                                                  fixbounds=False)
        # dprint("img_coords = %s" % str(img_coords))
        if img_coords != self.move_img_coords:
            self.resizeWorldCoordsFromImageCoords(*img_coords)
            self.draw()
            self.handleEventPostHook(ev)

    def handleMoveEvent(self, ev):
        coords = self.scroller.convertEventCoords(ev)
        img_coords = self.scroller.getImageCoords(coords[0],
                                                  coords[1],
                                                  fixbounds=False)
        # dprint("img_coords = %s" % str(img_coords))
        if img_coords != self.move_img_coords:
            self.moveWorldCoordsFromImageCoords(*img_coords)
            self.draw()
            self.handleEventPostHook(ev)

    def handleCursorChanges(self, ev):
        coords = self.scroller.convertEventCoords(ev)
        # dprint("mouse=%s world=%s" % (coords, self.world_coords))
        self.resize_index = None
        if self.isOnBorder(coords):
            self.resize_index = self.getBorderCursorIndex(coords)
            cursor = self.resize_cursors[self.resize_index]
        elif self.isInside(coords):
            cursor = self.move_cursor
        else:
            cursor = self.cursor
        self.scroller.setCursor(cursor)

    def finishEvent(self, ev):
        # dprint()
        self.scroller.ReleaseMouse()
        self.want_autoscroll = False

    def handleEventPostHook(self, ev):
        x0, y0 = self.start_img_coords
        x1, y1 = self.last_img_coords
        # normalize coords
        if x0 > x1:
            x0, x1 = x1, x0
        if y0 > y1:
            y0, y1 = y1, y0
        if self.event_type == "move":
            rbev = RubberBandMotionEvent(upperLeftImageCoords = (x0, y0),
                                         lowerRightImageCoords = (x1, y1))
        else:
            rbev = RubberBandSizeEvent(upperLeftImageCoords = (x0, y0),
                                       lowerRightImageCoords = (x1, y1))
        wx.PostEvent(self.scroller, rbev)

    def drawSelector(self, dc):
        xoff, yoff = self.getViewOffset()
        x = self.world_coords[0] + xoff
        y = self.world_coords[1] + yoff
        w = self.world_coords[2]
        h = self.world_coords[3]

        # dprint("start=%s current=%s  xywh=%s" % (self.start_img_coords, self.last_img_coords, (x,y,w,h)))
        dc.SetPen(wx.Pen(wx.BLACK, 1))
        dc.DrawRectangle(x, y, w, h)
        dc.SetPen(wx.Pen(wx.WHITE, 1, wx.DOT))
        dc.DrawRectangle(x, y, w, h)

    def isOnBorder(self, coords):
        """Are the world coordinates on the selection border?

        Return true if the world coordinates specified are on or
        within a tolerance of the selecton border.
        """
        # dprint(self.world_coords)
        if self.world_coords is None:
            return False
        return pointOnBox(coords[0], coords[1], self.world_coords, self.border_sensitivity)

    def getBorderCursorIndex(self, coords):
        """Get resize cursor depending on position on border.

        Modified from wx.lib.mixins.rubberband: Return a position
        number in the range 0 .. 7 to indicate where on the box border
        the point is.  The layout is:

              0    1    2
              7         3
              6    5    4
        """
        x0, y0, x1, y1 = boxToExtent(self.world_coords)
        x = coords[0]
        y = coords[1]
        t = self.border_sensitivity
        if x >= x0-t and x <= x0+t:
            # O, 7, or 6
            if y >= y0-t and y <= y0+t:
                index = 0
            elif y >= y1-t and y <= y1+t:
                index = 6
            else:
                index = 7
        elif x >= x1-t and x <= x1+t:
            # 2, 3, or 4
            if y >= y0-t and y <= y0+t:
                index = 2
            elif y >= y1-t and y <= y1+t:
                index = 4
            else:
                index = 3
        elif y >= y0-t and y <= y0+t:
            index = 1
        else:
            index = 5
        return index

    def isInside(self, coords):
        """Are the world coordinates on the selection border?

        Return true if the world coordinates specified are on or
        within a tolerance of the selecton border.
        """
        # dprint(self.world_coords)
        if self.world_coords is None:
            return False
        return pointInBox(coords[0], coords[1], self.world_coords)

    def recalc(self):
        zoom = self.scroller.zoom
        x, y = self.last_img_coords
        xi, yi = self.start_img_coords
        if x > xi:
            x0 = int(xi * zoom)
            x1 = int((x+1) * zoom)
        else:
            x0 = int(x * zoom)
            x1 = int((xi+1) * zoom)

        if y > yi:
            y0 = int(yi * zoom)
            y1 = int((y+1) * zoom)
        else:
            y0 = int(y * zoom)
            y1 = int((yi+1) * zoom)

        # (x0, y0) and (x1, y1) always point within the pixel, so we
        # need to expand the area so that the highlighted area is
        # outside the selection region.
        self.world_coords = (x0 - 1, y0 - 1, x1 - x0 + 2, y1 - y0 + 2)

    def normalizeImageCoords(self):
        x0, y0 = self.start_img_coords
        x1, y1 = self.last_img_coords
        self.normalizeNewImageCoords(x0, y0, x1, y1)

    def normalizeNewImageCoords(self, x0, y0, x1, y1):
        if x0 > x1:
            x0, x1 = x1, x0
        if y0 > y1:
            y0, y1 = y1, y0
        self.start_img_coords = (x0, y0)
        self.last_img_coords = (x1, y1)

    def setWorldCoordsFromImageCoords(self, x, y):
        self.last_img_coords = (x, y)
        self.recalc()

    def moveWorldCoordsFromImageCoords(self, x, y):
        dx = x - self.move_img_coords[0]
        dy = y - self.move_img_coords[1]
        self.move_img_coords = (x, y)
        x0, y0 = self.start_img_coords
        x1, y1 = self.last_img_coords
        if x0 + dx < 0:
            dx = -x0
        elif x1 + dx >= self.scroller.img.GetWidth():
            dx = self.scroller.img.GetWidth() - x1 - 1
        if y0 + dy < 0:
            dy = -y0
        elif y1 + dy >= self.scroller.img.GetHeight():
            dy = self.scroller.img.GetHeight() - y1 - 1
        self.start_img_coords = (x0 + dx, y0 + dy)
        self.last_img_coords = (x1 + dx, y1 + dy)
        self.recalc()

    def resizeWorldCoordsFromImageCoords(self, x, y):
        dx = x - self.move_img_coords[0]
        dy = y - self.move_img_coords[1]
        self.move_img_coords = self.scroller.getBoundedCoords(x, y)
        index = self.resize_index
        if index == 1 or index == 5:
            dx = 0
        elif index == 3 or index == 7:
            dy = 0
        x0, y0 = self.start_img_coords
        x1, y1 = self.last_img_coords
        if index >= 1 and index <= 3:
            # fix lower left corner
            x1 += dx
            y0 += dy
        elif index == 4:
            # fix upper left corner
            x1 += dx
            y1 += dy
        elif index >=5 and index <=7:
            # fix upper right corner:
            x0 += dx
            y1 += dy
        else:
            # fix upper left corner
            x0 += dx
            y0 += dy
        self.start_img_coords = self.scroller.getBoundedCoords(x0, y0)
        self.last_img_coords = self.scroller.getBoundedCoords(x1, y1)
        self.recalc()

    def getSelectedBox(self):
        """Return selected box in image coords.

        Returns an (x, y, w, h) tuple of the selected rectangle.
        """
        x0, y0 = self.start_img_coords
        x1, y1 = self.last_img_coords
        if x0 > x1:
            x0, x1 = x1, x0
        if y0 > y1:
            y0, y1 = y1, y0
        return (x0, y0, x1 - x0 + 1, y1 - y0 + 1)

    def setSelection(self, x0, y0, x1, y1):
        """Programmatically set the selection rectangle.
        """
        self.start_img_coords = self.scroller.getBoundedCoords(x0, y0)
        self.last_img_coords = self.scroller.getBoundedCoords(x1, y1)
        self.recalc()
        self.draw()


class ImageScroller(wx.ScrolledWindow):
    dbg_call_seq = 0

    def __init__(self, parent, task, selector=Crosshair):
        wx.ScrolledWindow.__init__(self, parent, -1)
        self.task = task
        self.editor = None

        # Settings
        self.background_color = wx.Colour(160, 160, 160)
        self.use_checkerboard = True
        self.checkerboard_box_size = 8
        self.checkerboard_color = wx.Colour(96, 96, 96)
        self.max_zoom = 16.0
        self.min_zoom = 0.0625

        # internal storage
        self.orig_img = None
        self.img = None
        self.scaled_bmp = None
        self.width = 0
        self.height = 0
        self.zoom = 1.0
        self.crop = None

        # hacks
        self.just_scrolled = False

        # cursors
        self.default_cursor = wx.CURSOR_ARROW
        self.save_cursor = None

        self.Bind(wx.EVT_PAINT, self.OnPaint)

        # selectors and related storage
        self.use_selector = None
        self.selector = None
        self.selector_event_callback = None
        self.overlay = wx.Overlay()
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouseEvent)

        self.setSelector(selector)

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.document = editor.document
            self.set_data()

    def set_data(self):
        img = wx.EmptyImage(1,1)
        if not img.LoadMimeStream(self.document.bytestream, self.document.metadata.mime):
            #raise TypeError("Bad image -- either it really isn't an image, or wxPython doesn't support the image format.")
            img = wx.EmptyImage(1,1)
        self.setImage(img)

    def zoomIn(self, zoom=2):
        self.zoom *= zoom
        if self.zoom > self.max_zoom:
            self.zoom = self.max_zoom
        self._scaleImage()

    def zoomOut(self, zoom=2):
        self.zoom /= zoom
        if self.zoom < self.min_zoom:
            self.zoom = self.min_zoom
        self._scaleImage()

    def _clearBackground(self, dc, w, h):
        dc.SetBackground(wx.Brush(self.background_color))
        dc.Clear()

    def _checkerboardBackground(self, dc, w, h):
        # draw checkerboard for transparent background
        box = self.checkerboard_box_size
        y = 0
        while y < h:
            #dprint("y=%d, y/box=%d" % (y, (y/box)%2))
            x = box * ((y/box)%2)
            while x < w:
                dc.SetPen(wx.Pen(self.checkerboard_color))
                dc.SetBrush(wx.Brush(self.checkerboard_color))
                dc.DrawRectangle(x, y, box, box)
                #dprint("draw: xywh=%s" % ((x, y, box, box),))
                x += box*2
            y += box

    def _drawBackground(self, dc, w, h):
        self._clearBackground(dc, w, h)
        if self.use_checkerboard:
            self._checkerboardBackground(dc, w, h)

    def inOrigImage(self, x, y):
        if x>=0 and x<self.orig_img.GetWidth() and y>=0 and y<self.orig_img.GetHeight():
            return True
        return False

    def _getCroppedImage(self):
        """Returns cropped image.

        Creates and returns a new image if there is a cropping
        specified, otherwise just returns the original image
        unchanged.
        """
        if self.crop is not None and isinstance(self.crop, tuple):
            if self.inOrigImage(self.crop[0], self.crop[1]) and self.inOrigImage(self.crop[0] + self.crop[2] - 1, self.crop[1] + self.crop[3] - 1):
                return self.orig_img.GetSubImage(self.crop)
            else:
                print("trying to crop outside of image: %s" % str(self.crop))
        return self.orig_img

    def _scaleImage(self):
        """Creates new image at specified zoom factor.

        Creates a new image that is to be used as the background of
        the scrolled bitmap.  Currently, actually creates the entire
        image, which could lead to memory problems if the image is
        really huge and the zoom factor is large.
        """
        if self.orig_img is not None:
            self.img = self._getCroppedImage()
            w = int(self.img.GetWidth() * self.zoom)
            h = int(self.img.GetHeight() * self.zoom)
            dc = wx.MemoryDC()
            self.scaled_bmp = wx.EmptyBitmap(w, h)
            dc.SelectObject(self.scaled_bmp)
            self._drawBackground(dc, w, h)

            # For very large bitmaps, the memory consumption of the
            # BitmapFromImage call can cause memory errors.  So, here
            # it breaks up the source image into chunks and only calls
            # BitmapFromImage on the chunk.
            ydest = 0
            source_step = 256
            for ysource in range(0, self.img.GetHeight(), source_step):
                if ysource + source_step > self.img.GetHeight():
                    hsource = self.img.GetHeight() - ysource
                else:
                    hsource = source_step
                hdest = hsource * self.zoom
                crop = [0, ysource, self.img.GetWidth(), hsource]
                #dprint(crop)
                subimg = self.orig_img.GetSubImage(crop)
                bmp = wx.BitmapFromImage(subimg.Scale(w, hdest))
                dc.DrawBitmap(bmp, 0, ydest, True)
                ydest += hdest
            self.width = self.scaled_bmp.GetWidth()
            self.height = self.scaled_bmp.GetHeight()
        else:
            self.width = 10
            self.height = 10
        self.SetVirtualSize((self.width, self.height))
        rate = int(self.zoom)
        if rate < 1:
            rate = 1
        self.SetScrollRate(rate, rate)
        if self.selector:
            self.selector.recalc()
        self.Refresh()

    def setImage(self, img=None, zoom=None, rot=None,
                 vmirror=False, hmirror=False, crop=None):
        """Sets the control to contain a new image.

        Main user interaction with this control -- makes the control
        display a new image, discarding the old image.

        @param img: new wx.Image to display
        @param zoom: optional floating point zoom factor
        @param rot: not working yet
        @param vmirror: not working yet
        @param hmirror: not working yet
        @param crop: None for no cropping, or (x, y, w, h) tuple
        """
        if img is not None:
            # change the bitmap if specified
            self.bmp = None
            self.orig_img = img
        else:
            self.bmp = self.orig_img = None

        if zoom is not None:
            self.zoom = zoom

        self.crop = crop
        self.endActiveSelector()
        self._scaleImage()

    def setBitmap(self, bmp=None, zoom=None):
        """Set the control to display a new bitmap.

        Similar to setImage, but takes a wx.Bitmap instead of a
        wx.Image.
        """
        if bmp is not None:
            img = bmp.ConvertToImage()
            self.setImage(img, zoom)
        else:
            self.setImage(None, zoom)

    def setCrop(self, crop=None):
        """Set the image cropping.

        Crops the image to the specified size.  If crop is an (x, y,
        w, h) tuple, it crops the image as specified.  If crop is
        None, it restores the original image.
        """
        if crop is not None and self.crop is not None:
            # If we're cropping an image that's already cropped,
            # add the offsets to find the new subimage.
            x = crop[0] + self.crop[0]
            y = crop[1] + self.crop[1]
            crop = (x, y, crop[2], crop[3])
        self.crop = crop
        self.endActiveSelector()
        self._scaleImage()

    def copyToClipboard(self):
        """Copies current image to clipboard.

        Copies the current image, including scaling, zooming, etc. to
        the clipboard.
        """
        img = self._getCroppedImage()
        w = int(img.GetWidth() * self.zoom)
        h = int(img.GetHeight() * self.zoom)
        clip = wx.BitmapFromImage(img.Scale(w, h))
        bmpdo = wx.BitmapDataObject(clip)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(bmpdo)
            wx.TheClipboard.Close()

    def saveImage(self, filename):
        """Copies current image to clipboard.

        Copies the current image, including scaling, zooming, etc. to
        the clipboard.
        """
        handlers = {'.png': wx.BITMAP_TYPE_PNG,
                    '.jpg': wx.BITMAP_TYPE_JPEG,
                    }

        root, ext = os.path.splitext(filename)
        ext = ext.lower()
        if ext in handlers:
            try:
                status = self.scaled_bmp.SaveFile(filename, handlers[ext])
            except:
                status = False
            return status
        raise TypeError("Unknown image file extension %s" % ext)

    def convertEventCoords(self, ev):
        """Convert event coordinates to world coordinates.

        Convert the event coordinates to world coordinates by locating
        the offset of the scrolled window's viewport and adjusting the
        event coordinates.
        """
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        x = ev.GetX() + (xView * xDelta)
        y = ev.GetY() + (yView * yDelta)
        return (x, y)

    def getBoundedCoords(self, x, y):
        """Return image coordinates clipped to boundary of image."""

        if x<0: x=0
        elif x>=self.img.GetWidth(): x=self.img.GetWidth()-1
        if y<0: y=0
        elif y>=self.img.GetHeight(): y=self.img.GetHeight()-1
        return (x, y)

    def getImageCoords(self, x, y, fixbounds = True):
        """Convert scrolled window coordinates to image coordinates.

        Convert from the scrolled window coordinates (where (0,0) is
        the upper left corner when the window is scrolled to the
        top-leftmost position) to the corresponding point on the
        original (i.e. unzoomed, unrotated, uncropped) image.
        """
        x = int(x / self.zoom)
        y = int(y / self.zoom)
        if fixbounds:
            return self.getBoundedCoords(x, y)
        else:
            return (x, y)

    def isInBounds(self, x, y):
        """Check if world coordinates are on the image.

        Return True if the world coordinates lie on the image.
        """
        if self.img is None or x<0 or y<0 or x>=self.width or y>=self.height:
            return False
        return True

    def isEventInClientArea(self, ev):
        """Check if event is in the viewport.

        Return True if the event is within the visible viewport of the
        scrolled window.
        """
        size = self.GetClientSizeTuple()
        x = ev.GetX()
        y = ev.GetY()
        if x < 0 or x >= size[0] or y < 0 or y >= size[1]:
            return False
        return True

    def setCursor(self, cursor):
        """Set cursor for the window.

        A mild enhancement of the wx standard SetCursor that takes an
        integer id as well as a wx.StockCursor instance.
        """
        if isinstance(cursor, int):
            cursor = wx.StockCursor(cursor)
        self.SetCursor(cursor)

    def blankCursor(self, ev, coords=None):
        """Turn off cursor.

        Some selectors, like the crosshair selector, work better when
        the mouse cursor isn't obscuring the image.  This method is
        called by the selector itself to blank the cursor when over
        the image.
        """
        #dprint()
        if coords is None:
            coords = self.convertEventCoords(ev)
        if self.isInBounds(*coords) and self.isEventInClientArea(ev):
            if not self.save_cursor:
                self.save_cursor = True
                self.setCursor(wx.StockCursor(wx.CURSOR_BLANK))
        else:
            if self.save_cursor:
                self.setCursor(self.use_selector.cursor)
                self.save_cursor = False

    def getSelectorCoordsOnImage(self, with_cropped_offset=True):
        if self.selector:
            x, y = self.getImageCoords(*self.selector.world_coords)
            if self.crop is not None and with_cropped_offset:
                x += self.crop[0]
                y += self.crop[1]
            return (x, y)
        return None

    #### - Automatic scrolling  FIXME: doesn't work yet

    def autoScroll(self, ev):
        x = ev.GetX()
        y = ev.GetY()
        size = self.GetClientSizeTuple()
        if x < 0:
            dx = x
        elif x > size[0]:
            dx = x - size[0]
        else:
            dx = 0
        if y < 0:
            dy = y
        elif y > size[1]:
            dy = y - size[1]
        else:
            dy = 0
        self.autoScrollCallback(dx, dy)

    def autoScrollCallback(self, dx, dy):
        self.dbg_call_seq += 1
        #print("In autoScrollCallback %d: dx=%d dy=%d" % (self.dbg_call_seq, dx, dy))
        spx = self.GetScrollPos(wx.HORIZONTAL)
        spy = self.GetScrollPos(wx.VERTICAL)
        if self.selector:
            self.selector.cleanup()
        wx.CallAfter(self.Scroll, spx+dx, spy+dy)
        self.just_scrolled = True

    def setSelector(self, selector):
        if self.selector:
            self.selector.cleanup()
            self.selector = None
        self.use_selector = selector
        self.setCursor(self.use_selector.cursor)
        self.save_cursor = None

    def startSelector(self, ev=None):
        self.selector = self.use_selector(self, ev)

    def getActiveSelector(self):
        """Returns the currently active selector."""
        return self.selector

    def endActiveSelector(self):
        if self.selector:
            self.selector.cleanup()
            self.selector = None
        if self.save_cursor:
            self.setCursor(self.use_selector.cursor)
            self.save_cursor = None

    def OnMouseEvent(self, ev):
        """Driver to process mouse events.

        This is the main driver to process all mouse events that
        happen on the ImageScroller.  Once a selector is triggered by
        its event combination, it becomes the active selector and
        further mouse events are directed to its handler.
        """
        if self.img:
            inside = self.isEventInClientArea(ev)

            try:
                # First, process the event itself or start it up if it has
                # been triggered.
                if self.selector:
                    if not self.selector.processEvent(ev):
                        self.endActiveSelector()
                elif inside and self.use_selector.trigger(ev):
                    self.startSelector(ev)

                # Next, if we have a selector, process some user interface
                # side effects
                if self.selector:
                    if self.selector.want_autoscroll and not inside:
                        self.autoScroll(ev)
                    if self.selector.want_blank_cursor:
                        self.blankCursor(ev)
            except:
                # If something bad happens, make sure we return control of the
                # mouse pointer to the user
                if self.HasCapture():
                    self.ReleaseMouse()
                raise
        ev.Skip()

    def OnPaint(self, evt):
        self.dbg_call_seq += 1
        #print("In OnPaint %d" % self.dbg_call_seq)
        if self.scaled_bmp is not None:
            dc=wx.BufferedPaintDC(self, self.scaled_bmp, wx.BUFFER_VIRTUAL_AREA)
            # Note that the drawing actually happens when the dc goes
            # out of scope and is destroyed.
            self.OnPaintHook(evt, dc)

            # FIXME: This check for MSW is because it gets multiple onpaint
            # events, so make sure it's only called once for consecutive
            # repaints.  HACK!
            if self.selector and (wx.Platform != '__WXMSW__' or not self.just_scrolled):
                wx.CallAfter(self.selector.recalc_and_draw)
            #self.overlay.Reset()
        self.just_scrolled = False
        evt.Skip()

    def OnPaintHook(self, evt, dc):
        """Hook to draw any additional items onto the saved bitmap.

        Note that any changes made to the dc will be reflected in the
        saved bitmap, so subsequent times calling this function will
        continue to add new data to the image.
        """
        pass


if __name__ == '__main__':
    app   = wx.PySimpleApp()
    frame = wx.Frame(None, -1, title='ImageScroller Test', size=(500,500))
    frame.CreateStatusBar()

    # Add a panel that the rubberband will work on.
    panel = ImageScroller(frame)
    img = wx.ImageFromBitmap(wx.ArtProvider_GetBitmap(wx.ART_WARNING, wx.ART_OTHER, wx.Size(48, 48)))
    panel.setImage(img)

    # Add the callbacks
    def crosshairCallback(ev):
        x, y = ev.imageCoords
        frame.SetStatusText("x=%d y=%d" % (x, y))
    panel.Bind(EVT_CROSSHAIR_MOTION, crosshairCallback)

    def rubberbandMotionCallback(ev):
        x0, y0 = ev.upperLeftImageCoords
        x1, y1 = ev.lowerRightImageCoords
        frame.SetStatusText("moving rubberband: ul = (%d,%d) lr = (%d,%d)" % (x0, y0, x1, y1))
    panel.Bind(EVT_RUBBERBAND_MOTION, rubberbandMotionCallback)

    def rubberbandSizeCallback(ev):
        x0, y0 = ev.upperLeftImageCoords
        x1, y1 = ev.lowerRightImageCoords
        frame.SetStatusText("sizing rubberband: ul = (%d,%d) lr = (%d,%d)" % (x0, y0, x1, y1))
    panel.Bind(EVT_RUBBERBAND_SIZE, rubberbandSizeCallback)

    # Layout the frame
    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(panel,  1, wx.EXPAND | wx.ALL, 5)

    def buttonHandler(ev):
        id = ev.GetId()
        if id == 100:
            panel.zoomIn()
        elif id == 101:
            panel.zoomOut()
        elif id == 102:
            panel.setSelector(Crosshair)
        elif id == 103:
            panel.setSelector(RubberBand)
        elif id == 104:
            selector = panel.getActiveSelector()
            dprint("selector = %s" % selector)
            if isinstance(selector, RubberBand):
                box = selector.getSelectedBox()
                dprint("selected box = %s" % str(box))
                panel.setCrop(box)
        elif id == 105:
            panel.setCrop(None)
        elif id == 106:
            panel.setSelector(RubberBand)
            panel.startSelector()
            selector = panel.getActiveSelector()
            img = panel.orig_img
            selector.setSelection(0,0, img.GetWidth()/2, img.GetHeight()/2)
        elif id == 200:
            wildcard="*"
            dlg = wx.FileDialog(
                frame, message="Open File",
                defaultFile="", wildcard=wildcard, style=wx.OPEN)

            # Show the dialog and retrieve the user response. If it is the
            # OK response, process the data.
            if dlg.ShowModal() == wx.ID_OK:
                # This returns a Python list of files that were selected.
                paths = dlg.GetPaths()

                for path in paths:
                    dprint("open file %s:" % path)
                    fh = open(path, 'rb')
                    img = wx.EmptyImage()
                    if img.LoadStream(fh):
                        panel.setImage(img)
                    else:
                        dprint("Invalid image: %s" % path)
            # Destroy the dialog. Don't do this until you are done with it!
            # BAD things can happen otherwise!
            dlg.Destroy()
        elif id == 201:
            pass
        elif id == 202:
            panel.copyToClipboard()
    buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(buttonsizer, 0, wx.EXPAND | wx.ALL, 5)
    button = wx.Button(frame, 100, 'Zoom In')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 101, 'Zoom Out')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 102, 'Crosshair')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 103, 'Select')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)

    buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(buttonsizer, 0, wx.EXPAND | wx.ALL, 5)
    button = wx.Button(frame, 104, 'Crop')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 105, 'Uncrop')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 106, 'Select half')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)

    buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(buttonsizer, 0, wx.EXPAND | wx.ALL, 5)
    button = wx.Button(frame, 200, 'Load')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 202, 'Copy to Clipboard')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)

    frame.SetAutoLayout(1)
    frame.SetSizer(sizer)
    frame.Show(1)
    app.MainLoop()



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2011 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

from __future__ import (absolute_import, division, print_function)

import cgi
from six import PY2
from six.moves import cStringIO as StringIO
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from owslib.util import openURL, testXMLValue, nspath_eval, ServiceException
from owslib.etree import etree
from owslib.fgdc import Metadata
from owslib.iso import MD_Metadata
from owslib.ows import *
from owslib.fes import *
from owslib.crs import Crs
from owslib.feature import WebFeatureService_
from owslib.namespaces import Namespaces
from owslib.util import log

def get_namespaces():
    n = Namespaces()
    return n.get_namespaces(["gml","ogc","ows","wfs"])
namespaces = get_namespaces()

class WebFeatureService_1_1_0(WebFeatureService_):
    """Abstraction for OGC Web Feature Service (WFS).

    Implements IWebFeatureService.
    """
    def __new__(self,url, version, xml, parse_remote_metadata=False, timeout=30):
        """ overridden __new__ method

        @type url: string
        @param url: url of WFS capabilities document
        @type xml: string
        @param xml: elementtree object
        @type parse_remote_metadata: boolean
        @param parse_remote_metadata: whether to fully process MetadataURL elements
        @param timeout: time (in seconds) after which requests should timeout
        @return: initialized WebFeatureService_1_1_0 object
        """
        obj=object.__new__(self)
        obj.__init__(url, version, xml, parse_remote_metadata, timeout)
        return obj

    def __getitem__(self,name):
        ''' check contents dictionary to allow dict like access to service layers'''
        if name in self.__getattribute__('contents').keys():
            return self.__getattribute__('contents')[name]
        else:
            raise KeyError("No content named %s" % name)


    def __init__(self, url, version, xml=None, parse_remote_metadata=False, timeout=30):
        """Initialize."""
        self.url = url
        self.version = version
        self.timeout = timeout
        self._capabilities = None
        self.owscommon = OwsCommon('1.0.0')
        reader = WFSCapabilitiesReader(self.version)
        if xml:
            self._capabilities = reader.readString(xml)
        else:
            self._capabilities = reader.read(self.url)
        self._buildMetadata(parse_remote_metadata)

    def _buildMetadata(self, parse_remote_metadata=False):
        '''set up capabilities metadata objects: '''

        # ServiceIdentification
        val = self._capabilities.find(util.nspath_eval('ows:ServiceIdentification', namespaces))
        self.identification=ServiceIdentification(val,self.owscommon.namespace)
        # ServiceProvider
        val = self._capabilities.find(util.nspath_eval('ows:ServiceProvider', namespaces))
        self.provider=ServiceProvider(val,self.owscommon.namespace)
        # ServiceOperations metadata
        self.operations=[]
        for elem in self._capabilities.findall(util.nspath_eval('ows:OperationsMetadata/ows:Operation', namespaces)):
            self.operations.append(OperationsMetadata(elem, self.owscommon.namespace))

        # FilterCapabilities
        val = self._capabilities.find(util.nspath_eval('ogc:Filter_Capabilities', namespaces))
        self.filters=FilterCapabilities(val)

        #serviceContents metadata: our assumption is that services use a top-level
        #layer as a metadata organizer, nothing more.

        self.contents={}
        features = self._capabilities.findall(nspath_eval('wfs:FeatureTypeList/wfs:FeatureType', namespaces))
        for feature in features:
            cm=ContentMetadata(feature, parse_remote_metadata)
            self.contents[cm.id]=cm

        #exceptions
        self.exceptions = [f.text for f \
                in self._capabilities.findall('Capability/Exception/Format')]

    def getcapabilities(self):
        """Request and return capabilities document from the WFS as a
        file-like object.
        NOTE: this is effectively redundant now"""
        reader = WFSCapabilitiesReader(self.version)
        return openURL(reader.capabilities_url(self.url), timeout=self.timeout)

    def items(self):
        '''supports dict-like items() access'''
        items=[]
        for item in self.contents:
            items.append((item,self.contents[item]))
        return items

    def _makeStringIO(self, strval):
        """
        Helper method to make sure the StringIO being returned will work.

        Differences between Python 2.6/2.7/3.x mean we have a lot of cases to handle.
        """
        if PY2:
            return StringIO(strval)

        return StringIO(strval.decode())

    def getfeature(self, typename=None, filter=None, bbox=None, featureid=None,
                   featureversion=None, propertyname='*', maxfeatures=None,
                   srsname=None, outputFormat=None, method='Get',
                   startindex=None):
        """Request and return feature data as a file-like object.

        Parameters
        ----------
        typename : list
            List of typenames (string)
        filter : string
            XML-encoded OGC filter expression.
        bbox : tuple
            (left, bottom, right, top) in the feature type's coordinates.
        featureid : list
            List of unique feature ids (string)
        featureversion : string
            Default is most recent feature version.
        propertyname : list
            List of feature property names. '*' matches all.
        maxfeatures : int
            Maximum number of features to be returned.
        method : string
            Qualified name of the HTTP DCP method to use.
        srsname: string
            EPSG code to request the data in
        outputFormat: string (optional)
            Requested response format of the request.
        startindex: int (optional)
            Start position to return feature set (paging in combination with maxfeatures)

        There are 3 different modes of use

        1) typename and bbox (simple spatial query). It is assumed, that
            bbox coordinates are given *always* in the east,north order
        2) typename and filter (more expressive)
        3) featureid (direct access to known features)
        """
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('GetFeature').methods if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url
        request = {'service': 'WFS', 'version': self.version, 'request': 'GetFeature'}

        if not isinstance(typename, list):
            typename = [typename]

        if srsname is not None:
            # check, if desired SRS is supported by the service for this typename
            if typename is not None:
                # convert srsname string to Crs object found in GetCaps
                srsnameobj = self.getSRS(srsname, typename[0])
                if srsnameobj is not None:
                    request['srsname'] = srsnameobj.id
                else:
                    options = ", ".join(map(lambda x: x.id, self.contents[typename[0]].crsOptions))
                    raise ServiceException("SRSNAME %s not supported.  Options: %s" % (srsname, options))
            else:
                request['srsname'] = str(srsname)

        # check featureid
        if featureid:
            request['featureid'] = ','.join(featureid)

        # bbox
        elif bbox and typename:
            request["bbox"] = self.getBBOXKVP(bbox, typename)

        # or filter
        elif filter and typename:
            request['filter'] = str(filter)

        assert len(typename) > 0
        request['typename'] = ','.join(typename)

        if propertyname is not None:
            if not isinstance(propertyname, list):
                propertyname = [propertyname]
            request['propertyname'] = ','.join(propertyname)

        if featureversion is not None:
            request['featureversion'] = str(featureversion)
        if maxfeatures is not None:
            request['maxfeatures'] = str(maxfeatures)
        if startindex is not None:
            request['startindex'] = str(startindex)
        if outputFormat is not None:
            request["outputFormat"] = outputFormat

        data = urlencode(request)
        log.debug("Making request: %s?%s" % (base_url, data))
        u = openURL(base_url, data, method, timeout=self.timeout)

        # check for service exceptions, rewrap, and return
        # We're going to assume that anything with a content-length > 32k
        # is data. We'll check anything smaller.
        if 'Content-Length' in u.info():
            length = int(u.info()['Content-Length'])
            have_read = False
        else:
            data = u.read()
            have_read = True
            length = len(data)

        if length < 32000:
            if not have_read:
                data = u.read()

            try:
                tree = etree.fromstring(data)
            except BaseException:
                # Not XML
                return self._makeStringIO(data)
            else:
                if tree.tag == "{%s}ServiceExceptionReport" % namespaces["ogc"]:
                    se = tree.find(nspath_eval('ServiceException', namespaces["ogc"]))
                    raise ServiceException(str(se.text).strip())
                else:
                    return self._makeStringIO(data)
        else:
            if have_read:
                return self._makeStringIO(data)
            return u

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)



class ContentMetadata:
    """Abstraction for WFS metadata.

    Implements IMetadata.
    """

    def __init__(self, elem, parse_remote_metadata=False, timeout=30):
        """."""
        self.id = testXMLValue(elem.find(nspath_eval('wfs:Name', namespaces)))
        self.title = testXMLValue(elem.find(nspath_eval('wfs:Title', namespaces)))
        self.abstract = testXMLValue(elem.find(nspath_eval('wfs:Abstract', namespaces)))
        self.keywords = [f.text for f in elem.findall(nspath_eval('ows:Keywords/ows:Keyword', namespaces))]

        # bbox
        self.boundingBoxWGS84 = None
        b = BoundingBox(elem.find(nspath_eval('ows:WGS84BoundingBox', namespaces)), namespaces['ows'])
        if b is not None:
            self.boundingBoxWGS84 = (
                    float(b.minx), float(b.miny),
                    float(b.maxx), float(b.maxy),
                    )
        # crs options
        self.crsOptions = [Crs(srs.text) for srs in elem.findall(nspath_eval('wfs:OtherSRS', namespaces))]
        dsrs = testXMLValue(elem.find(nspath_eval('wfs:DefaultSRS', namespaces)))
        if dsrs is not None:  # first element is default srs
            self.crsOptions.insert(0, Crs(dsrs))

        # verbs
        self.verbOptions = [op.text for op in elem.findall(nspath_eval('wfs:Operations/wfs:Operation', namespaces))]

        # output formats
        self.outputFormats = [op.text for op in elem.findall(nspath_eval('wfs:OutputFormats/wfs:Format', namespaces))]

        # MetadataURLs
        self.metadataUrls = []
        for m in elem.findall(nspath_eval('wfs:MetadataURL', namespaces)):
            metadataUrl = {
                'type': testXMLValue(m.attrib['type'], attrib=True),
                'format': testXMLValue(m.find('Format')),
                'url': testXMLValue(m)
            }

            if metadataUrl['url'] is not None and parse_remote_metadata:  # download URL
                try:
                    content = openURL(metadataUrl['url'], timeout=timeout)
                    doc = etree.parse(content)
                    if metadataUrl['type'] is not None:
                        if metadataUrl['type'] == 'FGDC':
                            metadataUrl['metadata'] = Metadata(doc)
                        if metadataUrl['type'] in ['TC211', '19115', '19139']:
                            metadataUrl['metadata'] = MD_Metadata(doc)
                except Exception:
                    metadataUrl['metadata'] = None

            self.metadataUrls.append(metadataUrl)

        #others not used but needed for iContentMetadata harmonisation
        self.styles=None
        self.timepositions=None
        self.defaulttimeposition=None

class WFSCapabilitiesReader(object):
    """Read and parse capabilities document into a lxml.etree infoset
    """

    def __init__(self, version='1.0'):
        """Initialize"""
        self.version = version
        self._infoset = None

    def capabilities_url(self, service_url):
        """Return a capabilities url
        """
        qs = []
        if service_url.find('?') != -1:
            qs = cgi.parse_qsl(service_url.split('?')[1])

        params = [x[0] for x in qs]

        if 'service' not in params:
            qs.append(('service', 'WFS'))
        if 'request' not in params:
            qs.append(('request', 'GetCapabilities'))
        if 'version' not in params:
            qs.append(('version', self.version))

        urlqs = urlencode(tuple(qs))
        return service_url.split('?')[0] + '?' + urlqs

    def read(self, url, timeout=30):
        """Get and parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        Parameters
        ----------
        url : string
            The URL to the WFS capabilities document.
        timeout : number
            A timeout value (in seconds) for the request.
        """
        request = self.capabilities_url(url)
        u = openURL(request, timeout=timeout)
        return etree.fromstring(u.read())

    def readString(self, st):
        """Parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        string should be an XML capabilities document
        """
        if not isinstance(st, str) and not isinstance(st, bytes):
            raise ValueError("String must be of type string or bytes, not %s" % type(st))
        return etree.fromstring(st)




#------------------------------------------------------------------------------
#
#  Copyright (c) 2013, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#------------------------------------------------------------------------------

from traits.has_traits import HasTraits
from traits.trait_types import Bool, Enum, Instance, Str
from traitsui.handler import ModelView
from traitsui.view import View, Group
from traitsui.item import Item
from traitsui.editors.code_editor import CodeEditor


from traitsui.tests._tools import *


class CodeModel(HasTraits):
    code = Str('world domination code')


class CodeView(ModelView):
    model = Instance(CodeModel)
    show_line_numbers = Bool(True)
    style = Enum('simple', 'readonly')

    def default_traits_view(self):
        traits_view = View(
            Item('model.code',
                 editor=CodeEditor(show_line_numbers=self.show_line_numbers),
                 style=self.style)
        )
        return traits_view


@skip_if_not_qt4
def test_code_editor_show_line_numbers():
    """ CodeEditor should honor the `show_line_numbers` setting
    """
    def is_line_numbers_visible(ui):
        from pyface import qt
        txt_ctrl = ui.control.findChild(qt.QtGui.QPlainTextEdit)
        return txt_ctrl.line_number_widget.isVisible()

    def test_line_numbers_visibility(show=True):
        with store_exceptions_on_all_threads():
            code_model = CodeModel()
            code_view = CodeView(model=code_model,
                                 show_line_numbers=show)
            ui = code_view.edit_traits()
            nose.tools.assert_equal(is_line_numbers_visible(ui), show)
            ui.control.close()

    test_line_numbers_visibility(True)
    test_line_numbers_visibility(False)


@skip_if_not_qt4
def test_code_editor_readonly():
    """ Test readonly editor style for CodeEditor
    """
    from pyface import qt
    with store_exceptions_on_all_threads():
        code_model = CodeModel()
        code_view = CodeView(model=code_model,
                             style='readonly')
        ui = code_view.edit_traits()
        txt_ctrl = ui.control.findChild(qt.QtGui.QPlainTextEdit)
        nose.tools.assert_true(txt_ctrl.isReadOnly())

        # Test changing the object's text
        nose.tools.assert_equal(txt_ctrl.toPlainText(), code_model.code)
        code_model.code += 'some more code'
        nose.tools.assert_true(txt_ctrl.isReadOnly())
        nose.tools.assert_equal(txt_ctrl.toPlainText(), code_model.code)

        # Test changing the underlying object
        code_model2 = CodeModel(code=code_model.code * 2)
        code_view.model = code_model2
        nose.tools.assert_true(txt_ctrl.isReadOnly())
        nose.tools.assert_equal(txt_ctrl.toPlainText(), code_model.code)

        ui.control.close()


if __name__ == '__main__':
    nose.main()



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/04/2007
#
#-------------------------------------------------------------------------

""" Traits UI themed checkbox editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Instance, Str

from traitsui.ui_traits \
    import ATheme, Image, Position, Spacing

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

from themed_control \
    import ThemedControl

#-------------------------------------------------------------------------
#  '_ThemedCheckboxEditor' class:
#-------------------------------------------------------------------------


class _ThemedCheckboxEditor(Editor):
    """ Traits UI themed checkbox editor.
    """

    # The ThemedControl used for the checkbox:
    checkbox = Instance(ThemedControl)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create the checkbox and its control:
        item = self.item
        factory = self.factory
        label = self.string_value(factory.label or item.label)
        min_size = (0, 0)
        if factory.theme is not None:
            min_size = (80, 0)

        self.checkbox = checkbox = ThemedControl(**factory.get(
            'image', 'position', 'spacing', 'theme')).set(
            text=label,
            controller=self,
            min_size=min_size)
        self.control = checkbox.create_control(parent)

        # Set the tooltip:
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self.checkbox.state == 'hover':
            self._set_hover_theme()
        else:
            self._set_theme()

    #-- ThemedControl Event Handlers -----------------------------------------

    def normal_motion(self, x, y, event):
        self._set_hover_theme('hover')
        self.control.CaptureMouse()

    def hover_left_down(self, x, y, event):
        self.control.ReleaseMouse()
        self._set_hover_theme('down', not self.value)

    def hover_motion(self, x, y, event):
        if not self.checkbox.in_control(x, y):
            self.control.ReleaseMouse()
            self._set_theme('normal')

    def down_left_up(self, x, y, event):
        if self.checkbox.in_control(x, y):
            self.value = not self.value
            self.normal_motion(x, y, event)
        else:
            self._set_theme('normal')

    def down_motion(self, x, y, event):
        if not self.checkbox.in_control(x, y):
            self._set_theme()
        else:
            self._set_hover_theme(value=not self.value)

    #-- Private Methods ------------------------------------------------------

    def _set_theme(self, state=None, value=None):
        """ Sets the theme, image, offset and optional checkbox state to use for
            a specified checkbox state value.
        """
        if value is None:
            value = self.value

        factory = self.factory
        theme, image = factory.theme, factory.image
        if value:
            theme, image = factory.on_theme, factory.on_image

        n = (1 * value) * (theme is not None)
        self.checkbox.set(offset=(n, n),
                          theme=theme or factory.theme,
                          image=image or factory.image,
                          state=state or self.checkbox.state)

    def _set_hover_theme(self, state=None, value=None):
        """ Sets the theme, image, offset and optional checkbox state to use for
            a specified checkbox state value while in hover mode.
        """
        if value is None:
            value = self.value

        factory = self.factory
        theme, image = factory.hover_off_theme, factory.hover_off_image
        if value:
            theme = factory.hover_on_theme or factory.on_theme
            image = factory.hover_on_image or factory.on_image

        n = (1 * value) * (theme is not None)
        self.checkbox.set(offset=(n, n),
                          theme=theme or factory.theme,
                          image=image or factory.image,
                          state=state or self.checkbox.state)

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for themed checkbox editors:


class ThemedCheckboxEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _ThemedCheckboxEditor

    # The checkbox label:
    label = Str

    # The basic theme for the checkbox (i.e. the 'off' state):
    theme = ATheme

    # The optional 'on' state theme for the checkbox:
    on_theme = ATheme

    # The optional 'hover off' state theme for the checkbox:
    hover_off_theme = ATheme

    # The optional 'hover on' state theme for the checbox:
    hover_on_theme = ATheme

    # The optional image to display in the checkbox (i.e. the 'off' state):
    image = Image('cb_off')

    # The optional 'on' state image to display in the checkbox:
    on_image = Image('cb_on')

    # The optional 'hover off' state image to display in the checkbox:
    hover_off_image = Image('cb_hover_off')

    # The optional 'hover on' state image to display in the checkbox:
    hover_on_image = Image('cb_hover_on')

    # The position of the image relative to the text:
    position = Position

    # The amount of space between the image and the text:
    spacing = Spacing

#-------------------------------------------------------------------------
#  Helper function for creating themed checkboxes:
#-------------------------------------------------------------------------


def themed_checkbox_editor(style=None, show_checkbox=True, **traits):
    """ Simplifies creation of a ThemedCheckboxEditor by setting up the
        themes and images automatically based on the value of the *style* and
        *show_checkbox* arguments.
    """
    tce = ThemedCheckboxEditor(**traits)

    if not show_checkbox:
        tce.set(image=None,
                on_image=None,
                hover_off_image=None,
                hover_on_image=None)

    if isinstance(style, basestring):
        group = style[0:1].upper()
        if (len(group) == 0) or (group not in 'BCGJTY'):
            group = 'B'

        row = style[1:2].upper()
        all_rows = '0123456789ABCDEFGHIJKL'
        if (len(row) == 0) or (row not in all_rows):
            row = 'H'

        column = style[2:3].upper()
        all_columns = '12345789AB'
        if (len(column) == 0) or (column not in all_columns):
            column = '5'

        tce.theme = '@%s%s%s' % (group, row, column)

        if style[-1:] == '.':
            return tce

        alt_row = '44456349A78FFFGHEFKLIJ'[all_rows.index(row)]
        alt_column = '66666CCCCC'[all_columns.index(column)]

        tce.set(on_theme='@%s%s%s' % (group, alt_row, column),
                hover_on_theme='@%s%s%s' % (group, alt_row, alt_column),
                hover_off_theme='@%s%s%s' % (group, row, alt_column))

    return tce



import unittest

from fs.tempfs import TempFS
from fs.memoryfs import MemoryFS
from fs import utils

from six import b

class TestUtils(unittest.TestCase):
    
    def _make_fs(self, fs):
        fs.setcontents("f1", b("file 1"))
        fs.setcontents("f2", b("file 2"))
        fs.setcontents("f3", b("file 3"))
        fs.makedir("foo/bar", recursive=True)
        fs.setcontents("foo/bar/fruit", b("apple"))
        
    def _check_fs(self, fs):
        self.assert_(fs.isfile("f1"))
        self.assert_(fs.isfile("f2"))
        self.assert_(fs.isfile("f3"))
        self.assert_(fs.isdir("foo/bar"))
        self.assert_(fs.isfile("foo/bar/fruit"))
        self.assertEqual(fs.getcontents("f1", "rb"), b("file 1"))
        self.assertEqual(fs.getcontents("f2", "rb"), b("file 2"))
        self.assertEqual(fs.getcontents("f3", "rb"), b("file 3"))
        self.assertEqual(fs.getcontents("foo/bar/fruit", "rb"), b("apple"))        
        
    def test_copydir_root(self):
        """Test copydir from root"""
        fs1 = MemoryFS()
        self._make_fs(fs1)        
        fs2 = MemoryFS()
        utils.copydir(fs1, fs2)        
        self._check_fs(fs2)
                
        fs1 = TempFS()
        self._make_fs(fs1)        
        fs2 = TempFS()
        utils.copydir(fs1, fs2)        
        self._check_fs(fs2)
    
    def test_copydir_indir(self):
        """Test copydir in a directory"""        
        fs1 = MemoryFS()
        fs2 = MemoryFS()
        self._make_fs(fs1)        
        utils.copydir(fs1, (fs2, "copy"))        
        self._check_fs(fs2.opendir("copy"))

        fs1 = TempFS()
        fs2 = TempFS()
        self._make_fs(fs1)        
        utils.copydir(fs1, (fs2, "copy"))        
        self._check_fs(fs2.opendir("copy"))
    
    def test_movedir_indir(self):
        """Test movedir in a directory"""        
        fs1 = MemoryFS()
        fs2 = MemoryFS()
        fs1sub = fs1.makeopendir("from")
        self._make_fs(fs1sub)            
        utils.movedir((fs1, "from"), (fs2, "copy"))        
        self.assert_(not fs1.exists("from"))     
        self._check_fs(fs2.opendir("copy"))

        fs1 = TempFS()
        fs2 = TempFS()
        fs1sub = fs1.makeopendir("from")
        self._make_fs(fs1sub)            
        utils.movedir((fs1, "from"), (fs2, "copy"))
        self.assert_(not fs1.exists("from"))      
        self._check_fs(fs2.opendir("copy"))
        
    def test_movedir_root(self):
        """Test movedir to root dir"""        
        fs1 = MemoryFS()
        fs2 = MemoryFS()
        fs1sub = fs1.makeopendir("from")
        self._make_fs(fs1sub)            
        utils.movedir((fs1, "from"), fs2)
        self.assert_(not fs1.exists("from"))     
        self._check_fs(fs2)

        fs1 = TempFS()
        fs2 = TempFS()
        fs1sub = fs1.makeopendir("from")
        self._make_fs(fs1sub)            
        utils.movedir((fs1, "from"), fs2)
        self.assert_(not fs1.exists("from"))        
        self._check_fs(fs2)
        
    def test_remove_all(self):
        """Test remove_all function"""
        fs = TempFS()
        fs.setcontents("f1", b("file 1"))
        fs.setcontents("f2", b("file 2"))
        fs.setcontents("f3", b("file 3"))
        fs.makedir("foo/bar", recursive=True)
        fs.setcontents("foo/bar/fruit", b("apple"))        
        fs.setcontents("foo/baz", b("baz"))
        
        utils.remove_all(fs, "foo/bar")
        self.assert_(not fs.exists("foo/bar/fruit"))
        self.assert_(fs.exists("foo/bar"))
        self.assert_(fs.exists("foo/baz"))
        utils.remove_all(fs,  "")
        self.assert_(not fs.exists("foo/bar/fruit"))
        self.assert_(not fs.exists("foo/bar/baz"))
        self.assert_(not fs.exists("foo/baz"))
        self.assert_(not fs.exists("foo"))
        self.assert_(not fs.exists("f1"))
        self.assert_(fs.isdirempty('/'))
    
    



#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt

from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..api import HasTraits, Instance, Str


class Shared(HasTraits):
    s = Str('new instance of Shared')


class Foo(HasTraits):
    s = Str('new instance of Foo')
    shared = Instance(Shared)


class Bar(HasTraits):
    s = Str('new instance of Bar')
    foo = Instance(Foo)
    shared = Instance(Shared)


class Baz(HasTraits):
    s = Str('new instance of Baz')
    bar = Instance(Bar)
    shared = Instance(Shared)


class CopyTraitsBase(unittest.TestCase):
    """ Validate that copy_traits
    """
    __test__ = False

    def setUp(self):
        super(CopyTraitsBase, self).setUp()
        self.shared = Shared(s='shared')
        self.foo = Foo(shared=self.shared, s='foo')
        self.bar = Bar(shared=self.shared, foo=self.foo, s='bar')
        self.baz = Baz(shared=self.shared, bar=self.bar, s='baz')

        self.shared2 = Shared(s='shared2')
        self.foo2 = Foo(shared=self.shared2, s='foo2')
        self.bar2 = Bar(shared=self.shared2, foo=self.foo2, s='bar2')
        self.baz2 = Baz(shared=self.shared2, bar=self.bar2, s='baz2')

        return

    def set_shared_copy(self, value):
        """ Change the copy style for the 'shared' traits. """
        self.foo.base_trait('shared').copy = value
        self.bar.base_trait('shared').copy = value
        self.baz.base_trait('shared').copy = value


class TestCopyTraitsSetup(CopyTraitsBase):
    __test__ = True

    def setUp(self):
        super(TestCopyTraitsSetup, self).setUp()

    def test_setup(self):
        self.assertIs(self.foo, self.bar.foo)
        self.assertIs(self.bar, self.baz.bar)
        self.assertIs(self.foo.shared, self.shared)
        self.assertIs(self.bar.shared, self.shared)
        self.assertIs(self.baz.shared, self.shared)

        self.assertIs(self.foo2, self.bar2.foo)
        self.assertIs(self.bar2, self.baz2.bar)
        self.assertIs(self.foo2.shared, self.shared2)
        self.assertIs(self.bar2.shared, self.shared2)
        self.assertIs(self.baz2.shared, self.shared2)


class CopyTraits(object):

    def test_baz2_s(self):
        self.assertEqual(self.baz2.s, 'baz')
        self.assertEqual(self.baz2.s, self.baz.s)

    def test_baz2_bar_s(self):
        self.assertEqual(self.baz2.bar.s, 'bar')
        self.assertEqual(self.baz2.bar.s, self.baz.bar.s)

    def test_baz2_bar_foo_s(self):
        self.assertEqual(self.baz2.bar.foo.s, 'foo')
        self.assertEqual(self.baz2.bar.foo.s, self.baz.bar.foo.s)

    def test_baz2_shared_s(self):
        self.assertEqual(self.baz2.shared.s, 'shared')
        self.assertEqual(self.baz2.bar.shared.s, 'shared')
        self.assertEqual(self.baz2.bar.foo.shared.s, 'shared')

    def test_baz2_bar(self):
        # First hand Instance trait is different and
        # is not the same object as the source.

        self.assertIsNot(self.baz2.bar, None)
        self.assertIsNot(self.baz2.bar, self.bar2)
        self.assertIsNot(self.baz2.bar, self.baz.bar)

    def test_baz2_bar_foo(self):
        # Second hand Instance trait is a different object and
        # is not the same object as the source.

        self.assertIsNot(self.baz2.bar.foo, None)
        self.assertIsNot(self.baz2.bar.foo, self.foo2)
        self.assertIsNot(self.baz2.bar.foo, self.baz.bar.foo)


class CopyTraitsSharedCopyNone(object):
    def test_baz2_shared(self):
        # First hand Instance trait is a different object and
        # is not the same object as the source.

        self.assertIsNot(self.baz2.shared, None)
        self.assertIsNot(self.baz2.shared, self.shared2)
        self.assertIsNot(self.baz2.shared, self.shared)

    def test_baz2_bar_shared(self):
        # Second hand Instance that was shared is a different object and
        # not the same object as the source and
        # not the same object as the new first hand instance that was the same.
        # I.e. There are now (at least) two copies of one original object.

        self.assertIsNot(self.baz2.bar.shared, None)
        self.assertIsNot(self.baz2.bar.shared, self.shared2)
        self.assertIsNot(self.baz2.bar.shared, self.shared)
        self.assertIsNot(self.baz2.bar.shared, self.baz2.shared)

    def test_baz2_bar_foo_shared(self):
        # Third hand Instance that was shared is a different object and
        # not the same object as the source and
        # not the same object as the new first hand instance that was the same.
        # I.e. There are now (at least) two copies of one original object.

        self.assertIsNot(self.baz2.bar.foo.shared, None)
        self.assertIsNot(self.baz2.bar.foo.shared, self.shared2)
        self.assertIsNot(self.baz2.bar.foo.shared, self.shared)
        self.assertIsNot(self.baz2.bar.foo.shared, self.baz2.shared)

    def test_baz2_bar_and_foo_shared(self):
        #
        # THE BEHAVIOR DEMONSTRATED BY THIS TEST CASE DOES NOT SEEM TO BE
        # CORRECT.
        #
        # Second and Third hand Instance object that was shared with first hand
        # instance are the same as each other but
        # Every reference to the same original object has been replace by
        # a reference to the same copy of the same source object except the
        # first hand reference which is a different copy.
        # I.e. The shared relationship has been fubarred by copy_traits: it's
        # not maintained, but not completely destroyed.
        self.assertIs(self.baz2.bar.shared, self.baz2.bar.foo.shared)
        self.assertIsNot(self.baz2.shared, self.baz2.bar.foo.shared)


class TestCopyTraitsSharedCopyNone(CopyTraits,
                                   CopyTraitsSharedCopyNone):
    __test__ = False

    def setUp(self):
        #super(TestCopyTraitsSharedCopyNone,self).setUp()

        # deep is the default value for Instance trait copy
        self.set_shared_copy('deep')
        return


class TestCopyTraitsCopyNotSpecified(CopyTraitsBase,
                                     TestCopyTraitsSharedCopyNone):
    __test__ = True

    def setUp(self):
#        super(TestCopyTraitsCopyNotSpecified,self).setUp()
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyNone.setUp(self)
        self.baz2.copy_traits(self.baz)
        return


class TestCopyTraitsCopyShallow(CopyTraitsBase, TestCopyTraitsSharedCopyNone):
    __test__ = True

    def setUp(self):
#        super(TestCopyTraitsCopyShallow,self).setUp()
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyNone.setUp(self)
        self.baz2.copy_traits(self.baz, copy='shallow')
        return


class TestCopyTraitsCopyDeep(CopyTraitsBase, TestCopyTraitsSharedCopyNone):
    __test__ = True

    def setUp(self):
#        super(TestCopyTraitsCopyDeep,self).setUp()
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyNone.setUp(self)
        self.baz2.copy_traits(self.baz, copy='deep')
        return


class CopyTraitsSharedCopyRef(object):
    def test_baz2_shared(self):
        # First hand Instance trait is a different object and
        # is the same object as the source.

        self.assertIsNot(self.baz2.shared, None)
        self.assertIsNot(self.baz2.shared, self.shared2)
        self.assertIs(self.baz2.shared, self.shared)

    def test_baz2_bar_shared(self):
        self.assertIsNot(self.baz2.bar.shared, None)
        self.assertIsNot(self.baz2.bar.shared, self.shared2)
        self.assertIs(self.baz2.bar.shared, self.shared)
        self.assertIs(self.baz2.bar.shared, self.baz2.shared)

    def test_baz2_bar_foo_shared(self):
        self.assertIsNot(self.baz2.bar.foo.shared, None)
        self.assertIsNot(self.baz2.bar.foo.shared, self.shared2)
        self.assertIs(self.baz2.bar.foo.shared, self.shared)
        self.assertIs(self.baz2.bar.foo.shared, self.baz2.shared)

    def test_baz2_bar_and_foo_shared(self):
        self.assertIs(self.baz2.bar.shared, self.baz2.bar.foo.shared)
        self.assertIs(self.baz2.shared, self.baz2.bar.foo.shared)


class TestCopyTraitsSharedCopyRef(CopyTraits,
                                  CopyTraitsSharedCopyRef):
    __test__ = False

    def setUp(self):
        #super(TestCopyTraitsSharedCopyRef,self).setUp()
        self.set_shared_copy('ref')
        return


# The next three tests demonstrate that a 'ref' trait is always copied as a
# reference regardless of the copy argument to copy_traits.  That is, shallow
# and deep are indistinguishable.
class TestCopyTraitsCopyNotSpecifiedSharedRef(CopyTraitsBase,
                                              TestCopyTraitsSharedCopyRef):
    __test__ = True

    def setUp(self):
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyRef.setUp(self)
        self.baz2.copy_traits(self.baz)
        return


class TestCopyTraitsCopyShallowSharedRef(CopyTraitsBase,
                                         TestCopyTraitsSharedCopyRef):
    __test__ = True

    def setUp(self):
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyRef.setUp(self)
        self.baz2.copy_traits(self.baz, copy='shallow')
        return


class TestCopyTraitsCopyDeepSharedRef(CopyTraitsBase,
                                      TestCopyTraitsSharedCopyRef):
    __test__ = True

    def setUp(self):
        CopyTraitsBase.setUp(self)
        TestCopyTraitsSharedCopyRef.setUp(self)
        self.baz2.copy_traits(self.baz, copy='deep')
        return
### EOF



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various text editors for the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#------------------------------------------------------------------------------

import wx

from traits.api \
    import TraitError

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.text_editor file.
from traitsui.editors.text_editor \
    import ToolkitEditorFactory, evaluate_trait

from editor \
    import Editor

from editor_factory \
    import ReadonlyEditor as BaseReadonlyEditor

from constants \
    import OKColor

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Readonly text editor with view state colors:
HoverColor = wx.LIGHT_GREY
DownColor = wx.WHITE

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style text editor, which displays a text field.
    """

    # Flag for window styles:
    base_style = 0

    # Background color when input is OK:
    ok_color = OKColor

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Function used to evaluate textual user input:
    evaluate = evaluate_trait

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        style = self.base_style
        self.evaluate = factory.evaluate
        self.sync_value(factory.evaluate_name, 'evaluate', 'from')

        if (not factory.multi_line) or factory.password:
            style &= ~wx.TE_MULTILINE

        if factory.password:
            style |= wx.TE_PASSWORD

        multi_line = ((style & wx.TE_MULTILINE) != 0)
        if multi_line:
            self.scrollable = True

        if factory.enter_set and (not multi_line):
            control = wx.TextCtrl(parent, -1, self.str_value,
                                  style=style | wx.TE_PROCESS_ENTER)
            wx.EVT_TEXT_ENTER(parent, control.GetId(), self.update_object)
        else:
            control = wx.TextCtrl(parent, -1, self.str_value, style=style)

        wx.EVT_KILL_FOCUS(control, self.update_object)

        if factory.auto_set:
            wx.EVT_TEXT(parent, control.GetId(), self.update_object)

        self.control = control
        self.set_error_state(False)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Handles the user entering input data in the edit control:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user entering input data in the edit control.
        """
        if isinstance(event, wx.FocusEvent):
            # Ensure that the base class' focus event handlers are run, some
            # built-in behavior may break on some platforms otherwise.
            event.Skip()

        if (not self._no_update) and (self.control is not None):
            try:
                self.value = self._get_user_value()

                if self._error is not None:
                    self._error = None
                    self.ui.errors -= 1

                self.set_error_state(False)

            except TraitError as excp:
                pass

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        user_value = self._get_user_value()
        try:
            unequal = bool(user_value != self.value)
        except ValueError:
            # This might be a numpy array.
            unequal = True

        if unequal:
            self._no_update = True
            self.control.SetValue(self.str_value)
            self._no_update = False

        if self._error is not None:
            self._error = None
            self.ui.errors -= 1
            self.set_error_state(False)

    #-------------------------------------------------------------------------
    #  Gets the actual value corresponding to what the user typed:
    #-------------------------------------------------------------------------

    def _get_user_value(self):
        """ Gets the actual value corresponding to what the user typed.
        """
        value = self.control.GetValue()
        try:
            value = self.evaluate(value)
        except:
            pass

        try:
            ret = self.factory.mapping.get(value, value)
        except TypeError:
            # The value is probably not hashable:
            ret = value

        return ret

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        if self._error is None:
            self._error = True
            self.ui.errors += 1

        self.set_error_state(True)

    #-------------------------------------------------------------------------
    #  Returns whether or not the editor is in an error state:
    #-------------------------------------------------------------------------

    def in_error_state(self):
        """ Returns whether or not the editor is in an error state.
        """
        return (self.invalid or self._error)

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(SimpleEditor):
    """ Custom style of text editor, which displays a multi-line text field.
    """

    # Flag for window style. This value overrides the default.
    base_style = wx.TE_MULTILINE

#-------------------------------------------------------------------------
#  'ReadonlyEditor' class:
#-------------------------------------------------------------------------


class ReadonlyEditor(BaseReadonlyEditor):
    """ Read-only style of text editor, which displays a read-only text field.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super(ReadonlyEditor, self).init(parent)

        if self.factory.view is not None:
            control = self.control
            wx.EVT_ENTER_WINDOW(control, self._enter_window)
            wx.EVT_LEAVE_WINDOW(control, self._leave_window)
            wx.EVT_LEFT_DOWN(control, self._left_down)
            wx.EVT_LEFT_UP(control, self._left_up)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        control = self.control
        new_value = self.str_value

        if hasattr(self.factory, 'password') and self.factory.password:
            new_value = '*' * len(new_value)

        if (self.item.resizable is True) or (self.item.height != -1.0):
            if control.GetValue() != new_value:
                control.SetValue(new_value)
                control.SetInsertionPointEnd()

        elif control.GetLabel() != new_value:
            control.SetLabel(new_value)

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self.factory.view is not None:
            control = self.control
            wx.EVT_ENTER_WINDOW(control, None)
            wx.EVT_LEAVE_WINDOW(control, None)
            wx.EVT_LEFT_DOWN(control, None)
            wx.EVT_LEFT_UP(control, None)

        super(ReadonlyEditor, self).dispose()

    #-- Private Methods ------------------------------------------------------

    def _set_color(self):
        control = self.control
        if not self._in_window:
            color = control.GetParent().GetBackgroundColour()
        elif self._down:
            color = DownColor
        else:
            color = HoverColor

        control.SetBackgroundColour(color)
        control.Refresh()

    #-- wxPython Event Handlers ----------------------------------------------

    def _enter_window(self, event):
        self._in_window = True
        self._set_color()

    def _leave_window(self, event):
        self._in_window = False
        self._set_color()

    def _left_down(self, event):
        self.control.CaptureMouse()
        self._down = True
        self._set_color()

    def _left_up(self, event):
        self._set_color()
        if not self._down:
            return

        self.control.ReleaseMouse()
        self._down = False

        if self._in_window:
            self.object.edit_traits(view=self.factory.view,
                                    parent=self.control)

TextEditor = SimpleEditor



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, Callable, HasTraits

# Local imports.
from bind_event import BindEvent
from package_globals import get_script_manager
from scriptable_type import make_object_scriptable


class FactoryWrapper(HasTraits):
    """The FactoryWrapper class wraps a factory for an object."""

    #### 'FactoryWrapper' interface ###########################################

    # The optional object that defines the scripting API.
    api = Any

    # The object factory.
    factory = Callable

    # The optional attribute include list.
    includes = Any

    # The optional attribute exclude list.
    excludes = Any

    ###########################################################################
    # 'FactoryWrapper' interface.
    ###########################################################################

    def create_scriptable_object(self, name):
        """Invoke the factory to create the object then make it scriptable."""

        obj = self.factory()

        sm = get_script_manager()
        sm.bind_event = BindEvent(name=name, obj=obj)

        make_object_scriptable(obj, self.api, self.includes, self.excludes)

        return obj


class _LazyNode(object):
    """The _LazyNode class implements a node in a lazy namespace that will
    automatically invoke a factory if one is referenced."""

    def __getattribute__(self, name):

        value = super(_LazyNode, self).__getattribute__(name)

        if isinstance(value, FactoryWrapper):
            value = value.create_scriptable_object(name)
            setattr(self, name, value)

        return value


class LazyNamespace(dict):
    """The LazyNamespace class implements a lazy namespace that will
    automatically invoke a factory if one is referenced."""

    def __getitem__(self, name):

        value = super(LazyNamespace, self).__getitem__(name)

        if isinstance(value, FactoryWrapper):
            value = value.create_scriptable_object(name)
            self[name] = value

        return value


def add_to_namespace(so, name, nspace):
    """Add a named scriptable object (or a factory for one) to a lazy
    namespace.  If a name is a dotted name then intermediary nodes in the
    namespace are created as required."""

    def save_obj(obj, name):

        if isinstance(nspace, LazyNamespace):
            nspace[name] = obj
        else:
            setattr(nspace, name, obj)

    parts = name.split('.')

    for part in parts[:-1]:
        if isinstance(nspace, LazyNamespace):
            try:
                next_nspace = nspace[part]
            except KeyError:
                next_nspace = _LazyNode()
        elif isinstance(nspace, _LazyNode):
            try:
                next_nspace = nspace[part]
            except KeyError:
                next_nspace = _LazyNode()
        else:
            raise NameError

        save_obj(next_nspace, part)
        nspace = next_nspace

    if not isinstance(nspace, (LazyNamespace, _LazyNode)):
        raise NameError

    save_obj(so, parts[-1])
